import os
import tempfile
from datetime import datetime

import absl.flags as flags
import ml_collections
import numpy as np
import wandb
from PIL import Image, ImageEnhance


class CsvLogger:
    """CSV logger for logging metrics to a CSV file."""

    def __init__(self, path):
        self.path = path
        self.header = None
        self.file = None
        self.disallowed_types = (wandb.Image, wandb.Video, wandb.Histogram)

    def log(self, row, step):
        row['step'] = step
        if self.file is None:
            self.file = open(self.path, 'w')
            if self.header is None:
                self.header = [k for k, v in row.items() if not isinstance(v, self.disallowed_types)]
                self.file.write(','.join(self.header) + '\n')
            filtered_row = {k: v for k, v in row.items() if not isinstance(v, self.disallowed_types)}
            self.file.write(','.join([str(filtered_row.get(k, '')) for k in self.header]) + '\n')
        else:
            filtered_row = {k: v for k, v in row.items() if not isinstance(v, self.disallowed_types)}
            self.file.write(','.join([str(filtered_row.get(k, '')) for k in self.header]) + '\n')
        self.file.flush()

    def close(self):
        if self.file is not None:
            self.file.close()


def get_exp_name(seed):
    """Return the experiment name."""
    exp_name = ''
    exp_name += f'sd{seed:03d}_'
    if 'SLURM_JOB_ID' in os.environ:
        exp_name += f's_{os.environ["SLURM_JOB_ID"]}.'
    if 'SLURM_PROCID' in os.environ:
        exp_name += f'{os.environ["SLURM_PROCID"]}.'
    exp_name += f'{datetime.now().strftime("%Y%m%d_%H%M%S")}'

    return exp_name


def get_flag_dict():
    """Return the dictionary of flags."""
    flag_dict = {k: getattr(flags.FLAGS, k) for k in flags.FLAGS if '.' not in k}
    for k in flag_dict:
        if isinstance(flag_dict[k], ml_collections.ConfigDict):
            flag_dict[k] = flag_dict[k].to_dict()
    return flag_dict


def setup_wandb(
    entity=None,
    project='project',
    group=None,
    name=None,
    mode='online',
):
    """Set up Weights & Biases for logging."""
    wandb_output_dir = tempfile.mkdtemp()
    tags = [group] if group is not None else None

    init_kwargs = dict(
        config=get_flag_dict(),
        project=project,
        entity=entity,
        tags=tags,
        group=group,
        dir=wandb_output_dir,
        name=name,
        settings=wandb.Settings(
            start_method='thread',
            _disable_stats=False,
        ),
        mode=mode,
        save_code=True,
    )

    run = wandb.init(**init_kwargs)

    return run


def reshape_video(v, n_cols=None):
    """Helper function to reshape videos."""
    if v.ndim == 4:
        v = v[None,]

    _, t, h, w, c = v.shape

    if n_cols is None:
        # Set n_cols to the square root of the number of videos.
        n_cols = np.ceil(np.sqrt(v.shape[0])).astype(int)
    if v.shape[0] % n_cols != 0:
        len_addition = n_cols - v.shape[0] % n_cols
        v = np.concatenate((v, np.zeros(shape=(len_addition, t, h, w, c))), axis=0)
    n_rows = v.shape[0] // n_cols

    v = np.reshape(v, newshape=(n_rows, n_cols, t, h, w, c))
    v = np.transpose(v, axes=(2, 5, 0, 3, 1, 4))
    v = np.reshape(v, newshape=(t, c, n_rows * h, n_cols * w))

    return v


def get_wandb_video(renders=None, n_cols=None, fps=15):
    """Return a Weights & Biases video.

    It takes a list of videos and reshapes them into a single video with the specified number of columns.

    Args:
        renders: List of videos. Each video should be a numpy array of shape (t, h, w, c).
        n_cols: Number of columns for the reshaped video. If None, it is set to the square root of the number of videos.
    """
    # Pad videos to the same length.
    max_length = max([len(render) for render in renders])
    for i, render in enumerate(renders):
        assert render.dtype == np.uint8

        # Decrease brightness of the padded frames.
        final_frame = render[-1]
        final_image = Image.fromarray(final_frame)
        enhancer = ImageEnhance.Brightness(final_image)
        final_image = enhancer.enhance(0.5)
        final_frame = np.array(final_image)

        pad = np.repeat(final_frame[np.newaxis, ...], max_length - len(render), axis=0)
        renders[i] = np.concatenate([render, pad], axis=0)

        # Add borders.
        renders[i] = np.pad(renders[i], ((0, 0), (1, 1), (1, 1), (0, 0)), mode='constant', constant_values=0)
    renders = np.array(renders)  # (n, t, h, w, c)

    renders = reshape_video(renders, n_cols)  # (t, c, nr * h, nc * w)

    return wandb.Video(renders, fps=fps, format='mp4')



import numpy as np


def relabel_dataset(env_name, env, dataset):
    """Relabel the dataset with rewards and masks based on the fixed task of the environment.
    This is useful for single-task variants of the environments.

    Args:
        env_name: Name of the environment.
        env: Environment.
        dataset: Dataset dictionary.
    """
    assert env.unwrapped._reward_task_id is not None, 'The environment is not in the single-task mode.'
    env.reset()  # Set the task.

    if 'maze' in env_name or 'soccer' in env_name:
        # Locomotion environments.
        qpos_xy_start_idx = 0
        qpos_ball_start_idx = 15
        goal_xy = env.unwrapped.cur_goal_xy
        goal_tol = env.unwrapped._goal_tol

        # Compute successes.
        if 'maze' in env_name:
            dists = np.linalg.norm(dataset['qpos'][:, qpos_xy_start_idx : qpos_xy_start_idx + 2] - goal_xy, axis=-1)
        else:
            dists = np.linalg.norm(dataset['qpos'][:, qpos_ball_start_idx : qpos_ball_start_idx + 2] - goal_xy, axis=-1)
        successes = (dists <= goal_tol).astype(np.float32)

        rewards = successes - 1.0
        masks = 1.0 - successes
    elif 'cube' in env_name or 'scene' in env_name or 'puzzle' in env_name:
        # Manipulation environments.
        qpos_obj_start_idx = 14
        qpos_cube_length = 7

        if 'cube' in env_name:
            num_cubes = env.unwrapped._num_cubes
            target_cube_xyzs = env.unwrapped._data.mocap_pos.copy()

            # Compute successes.
            cube_xyzs_list = []
            for i in range(num_cubes):
                cube_xyzs_list.append(
                    dataset['qpos'][
                        :, qpos_obj_start_idx + i * qpos_cube_length : qpos_obj_start_idx + i * qpos_cube_length + 3
                    ]
                )
            cube_xyzs = np.stack(cube_xyzs_list, axis=1)
            successes = np.linalg.norm(target_cube_xyzs - cube_xyzs, axis=-1) <= 0.04
        elif 'scene' in env_name:
            num_cubes = env.unwrapped._num_cubes
            num_buttons = env.unwrapped._num_buttons
            qpos_drawer_idx = qpos_obj_start_idx + num_cubes * qpos_cube_length + num_buttons
            qpos_window_idx = qpos_drawer_idx + 1
            target_cube_xyzs = env.unwrapped._data.mocap_pos.copy()
            target_button_states = env.unwrapped._target_button_states.copy()
            target_drawer_pos = env.unwrapped._target_drawer_pos
            target_window_pos = env.unwrapped._target_window_pos

            # Compute successes.
            cube_xyzs_list = []
            for i in range(num_cubes):
                cube_xyzs_list.append(
                    dataset['qpos'][
                        :, qpos_obj_start_idx + i * qpos_cube_length : qpos_obj_start_idx + i * qpos_cube_length + 3
                    ]
                )
            cube_xyzs = np.stack(cube_xyzs_list, axis=1)
            cube_successes = np.linalg.norm(target_cube_xyzs - cube_xyzs, axis=-1) <= 0.04
            button_successes = dataset['button_states'] == target_button_states
            drawer_success = np.abs(dataset['qpos'][:, qpos_drawer_idx] - target_drawer_pos) <= 0.04
            window_success = np.abs(dataset['qpos'][:, qpos_window_idx] - target_window_pos) <= 0.04
            successes = np.concatenate(
                [cube_successes, button_successes, drawer_success[:, None], window_success[:, None]], axis=-1
            )
        elif 'puzzle' in env_name:
            num_buttons = env.unwrapped._num_buttons
            target_button_states = env.unwrapped._target_button_states.copy()

            # Compute successes.
            successes = dataset['button_states'] == target_button_states

        rewards = successes.sum(axis=-1) - successes.shape[-1]
        masks = 1.0 - np.all(successes, axis=-1)
    else:
        raise ValueError(f'Unsupported environment: {env_name}')

    dataset['rewards'] = rewards.astype(np.float32)
    dataset['masks'] = masks.astype(np.float32)


def add_oracle_reps(env_name, env, dataset):
    """Add oracle goal representations to the dataset.

    Args:
        env_name: Name of the environment.
        env: Environment.
        dataset: Dataset dictionary.
    """
    if 'maze' in env_name or 'soccer' in env_name:
        # Locomotion environments.
        qpos_xy_start_idx = 0
        qpos_ball_start_idx = 15

        if 'maze' in env_name:
            oracle_reps = dataset['qpos'][:, qpos_xy_start_idx : qpos_xy_start_idx + 2]
        else:
            oracle_reps = dataset['qpos'][:, qpos_ball_start_idx : qpos_ball_start_idx + 2]
    elif 'cube' in env_name or 'scene' in env_name or 'puzzle' in env_name:
        # Manipulation environments.
        qpos_obj_start_idx = 14
        qpos_cube_length = 7
        xyz_center = np.array([0.425, 0.0, 0.0])
        xyz_scaler = 10.0
        drawer_scaler = 18.0
        window_scaler = 15.0

        if 'cube' in env_name:
            num_cubes = env.unwrapped._num_cubes

            cube_xyzs_list = []
            for i in range(num_cubes):
                cube_xyzs_list.append(
                    dataset['qpos'][
                        :, qpos_obj_start_idx + i * qpos_cube_length : qpos_obj_start_idx + i * qpos_cube_length + 3
                    ]
                )
            cube_xyzs = np.stack(cube_xyzs_list, axis=1)
            oracle_reps = ((cube_xyzs - xyz_center) * xyz_scaler).reshape(-1, num_cubes * 3)
        elif 'scene' in env_name:
            num_cubes = env.unwrapped._num_cubes
            num_buttons = env.unwrapped._num_buttons
            qpos_drawer_idx = qpos_obj_start_idx + num_cubes * qpos_cube_length + num_buttons
            qpos_window_idx = qpos_drawer_idx + 1

            cube_xyzs_list = []
            for i in range(num_cubes):
                cube_xyzs_list.append(
                    dataset['qpos'][
                        :, qpos_obj_start_idx + i * qpos_cube_length : qpos_obj_start_idx + i * qpos_cube_length + 3
                    ]
                )
            cube_xyzs = np.stack(cube_xyzs_list, axis=1)
            cube_reps = ((cube_xyzs - xyz_center) * xyz_scaler).reshape(-1, num_cubes * 3)
            button_reps = dataset['button_states'].copy()
            drawer_reps = dataset['qpos'][:, [qpos_drawer_idx]] * drawer_scaler
            window_reps = dataset['qpos'][:, [qpos_window_idx]] * window_scaler
            oracle_reps = np.concatenate([cube_reps, button_reps, drawer_reps, window_reps], axis=-1)
        elif 'puzzle' in env_name:
            oracle_reps = dataset['button_states'].copy()
    else:
        raise ValueError(f'Unsupported environment: {env_name}')

    dataset['oracle_reps'] = oracle_reps.astype(np.float32)



import mujoco
import numpy as np

PI = np.pi
PI_2 = 2 * np.pi


def angle_diff(q1: np.ndarray, q2: np.ndarray) -> np.ndarray:
    return np.mod(q1 - q2 + PI, PI_2) - PI


class DiffIKController:
    """Differential inverse kinematics controller."""

    def __init__(
        self,
        model: mujoco.MjModel,
        sites: list,
        qpos0: np.ndarray = None,
        damping_coeff: float = 1e-12,
        max_angle_change: float = np.radians(45),
    ):
        self._model = model
        self._data = mujoco.MjData(self._model)
        self._qp0 = qpos0
        self._max_angle_change = max_angle_change

        # Cache references.
        self._ns = len(sites)  # Number of sites.
        self._site_ids = np.asarray([self._model.site(s).id for s in sites])

        # Preallocate arrays.
        self._err = np.empty((self._ns, 6))
        self._site_quat = np.empty((self._ns, 4))
        self._site_quat_inv = np.empty((self._ns, 4))
        self._err_quat = np.empty((self._ns, 4))
        self._jac = np.empty((6 * self._ns, self._model.nv))
        self._damping = damping_coeff * np.eye(6 * self._ns)
        self._eye = np.eye(self._model.nv)

    def _forward_kinematics(self) -> None:
        """Minimal computation required for forward kinematics."""
        mujoco.mj_kinematics(self._model, self._data)
        mujoco.mj_comPos(self._model, self._data)  # Required for mj_jacSite.

    def _integrate(self, update: np.ndarray) -> None:
        """Integrate the joint velocities in-place."""
        mujoco.mj_integratePos(self._model, self._data.qpos, update, 1.0)

    def _compute_translational_error(self, pos: np.ndarray) -> None:
        """Compute the error between the desired and current site positions."""
        self._err[:, :3] = pos - self._data.site_xpos[self._site_ids]

    def _compute_rotational_error(self, quat: np.ndarray) -> None:
        """Compute the error between the desired and current site orientations."""
        for i, site_id in enumerate(self._site_ids):
            mujoco.mju_mat2Quat(self._site_quat[i], self._data.site_xmat[site_id])
            mujoco.mju_negQuat(self._site_quat_inv[i], self._site_quat[i])
            mujoco.mju_mulQuat(self._err_quat[i], quat[i], self._site_quat_inv[i])
            mujoco.mju_quat2Vel(self._err[i, 3:], self._err_quat[i], 1.0)

    def _compute_jacobian(self) -> None:
        """Update site end-effector Jacobians."""
        for i, site_id in enumerate(self._site_ids):
            jacp = self._jac[6 * i : 6 * i + 3]
            jacr = self._jac[6 * i + 3 : 6 * i + 6]
            mujoco.mj_jacSite(self._model, self._data, jacp, jacr, site_id)

    def _error_threshold_reached(self, pos_thresh: float, ori_thresh: float) -> bool:
        """Return True if position and rotation errors are below the thresholds."""
        pos_achieved = np.linalg.norm(self._err[:, :3]) <= pos_thresh
        ori_achieved = np.linalg.norm(self._err[:, 3:]) <= ori_thresh
        return pos_achieved and ori_achieved

    def _solve(self) -> np.ndarray:
        """Solve for joint velocities using damped least squares."""
        H = self._jac @ self._jac.T + self._damping
        x = self._jac.T @ np.linalg.solve(H, self._err.ravel())
        if self._qp0 is not None:
            jac_pinv = np.linalg.pinv(H)
            q_err = angle_diff(self._qp0, self._data.qpos)
            x += (self._eye - (self._jac.T @ jac_pinv) @ self._jac) @ q_err
        return x

    def _scale_update(self, update: np.ndarray) -> np.ndarray:
        """Scale down update so that the max allowable angle change is not exceeded."""
        update_max = np.max(np.abs(update))
        if update_max > self._max_angle_change:
            update *= self._max_angle_change / update_max
        return update

    def solve(
        self,
        pos: np.ndarray,
        quat: np.ndarray,
        curr_qpos: np.ndarray,
        max_iters: int = 20,
        pos_thresh: float = 1e-4,
        ori_thresh: float = 1e-4,
    ) -> np.ndarray:
        self._data.qpos = curr_qpos

        for _ in range(max_iters):
            self._forward_kinematics()

            self._compute_translational_error(np.atleast_2d(pos))
            self._compute_rotational_error(np.atleast_2d(quat))
            if self._error_threshold_reached(pos_thresh, ori_thresh):
                break

            self._compute_jacobian()
            update = self._scale_update(self._solve())
            self._integrate(update)

        return self._data.qpos.copy()



import glob
import json
import pathlib
from collections import defaultdict

import gymnasium
import numpy as np
from absl import app, flags
from agents import SACAgent
from tqdm import trange
from utils.evaluation import supply_rng
from utils.flax_utils import restore_agent

import ogbench.locomaze  # noqa

FLAGS = flags.FLAGS

flags.DEFINE_integer('seed', 0, 'Random seed.')
flags.DEFINE_string('env_name', 'antmaze-large-v0', 'Environment name.')
flags.DEFINE_string('dataset_type', 'navigate', 'Dataset type.')
flags.DEFINE_string('restore_path', 'experts/ant', 'Expert agent restore path.')
flags.DEFINE_integer('restore_epoch', 400000, 'Expert agent restore epoch.')
flags.DEFINE_string('save_path', None, 'Save path.')
flags.DEFINE_float('noise', 0.2, 'Gaussian action noise level.')
flags.DEFINE_integer('num_episodes', 1000, 'Number of episodes.')
flags.DEFINE_integer('max_episode_steps', 1001, 'Maximum number of steps in an episode.')


def main(_):
    assert FLAGS.dataset_type in ['path', 'navigate', 'stitch', 'explore']
    # 'path': Reach a single goal and stay there.
    # 'navigate': Repeatedly reach randomly sampled goals in a single episode.
    # 'stitch': Reach a nearby goal that is 4 cells away and stay there.
    # 'explore': Repeatedly follow random directions sampled every 10 steps.

    # Initialize environment.
    env = gymnasium.make(
        FLAGS.env_name,
        terminate_at_goal=False,
        max_episode_steps=FLAGS.max_episode_steps,
    )
    ob_dim = env.observation_space.shape[0]

    # Initialize oracle agent.
    if 'point' in FLAGS.env_name:

        def actor_fn(ob, temperature):
            return ob[-2:]
    else:
        # Load agent config.
        restore_path = FLAGS.restore_path
        candidates = glob.glob(restore_path)
        assert len(candidates) == 1, f'Found {len(candidates)} candidates: {candidates}'

        with open(candidates[0] + '/flags.json', 'r') as f:
            agent_config = json.load(f)['agent']

        # Load agent.
        agent = SACAgent.create(
            FLAGS.seed,
            np.zeros(ob_dim),
            env.action_space.sample(),
            agent_config,
        )
        agent = restore_agent(agent, FLAGS.restore_path, FLAGS.restore_epoch)
        actor_fn = supply_rng(agent.sample_actions, rng=agent.rng)

    # Store all empty cells and vertex cells.
    all_cells = []
    vertex_cells = []
    maze_map = env.unwrapped.maze_map
    for i in range(maze_map.shape[0]):
        for j in range(maze_map.shape[1]):
            if maze_map[i, j] == 0:
                all_cells.append((i, j))

                # Exclude hallway cells.
                if (
                    maze_map[i - 1, j] == 0
                    and maze_map[i + 1, j] == 0
                    and maze_map[i, j - 1] == 1
                    and maze_map[i, j + 1] == 1
                ):
                    continue
                if (
                    maze_map[i, j - 1] == 0
                    and maze_map[i, j + 1] == 0
                    and maze_map[i - 1, j] == 1
                    and maze_map[i + 1, j] == 1
                ):
                    continue

                vertex_cells.append((i, j))

    # Collect data.
    dataset = defaultdict(list)
    total_steps = 0
    total_train_steps = 0
    num_train_episodes = FLAGS.num_episodes
    num_val_episodes = FLAGS.num_episodes // 10
    for ep_idx in trange(num_train_episodes + num_val_episodes):
        if FLAGS.dataset_type in ['path', 'navigate', 'explore']:
            # Sample an initial state from all cells.
            init_ij = all_cells[np.random.randint(len(all_cells))]
            # Sample a goal state from vertex cells.
            goal_ij = vertex_cells[np.random.randint(len(vertex_cells))]
        elif FLAGS.dataset_type == 'stitch':
            # Sample an initial state from all cells.
            init_ij = all_cells[np.random.randint(len(all_cells))]

            # Perform BFS to find adjacent cells.
            adj_cells = []
            adj_steps = 4  # Target distance from the initial cell.
            bfs_map = maze_map.copy()
            for i in range(bfs_map.shape[0]):
                for j in range(bfs_map.shape[1]):
                    bfs_map[i][j] = -1
            bfs_map[init_ij[0], init_ij[1]] = 0
            queue = [init_ij]
            while len(queue) > 0:
                i, j = queue.pop(0)
                for di, dj in [(-1, 0), (0, -1), (1, 0), (0, 1)]:
                    ni, nj = i + di, j + dj
                    if (
                        0 <= ni < bfs_map.shape[0]
                        and 0 <= nj < bfs_map.shape[1]
                        and maze_map[ni, nj] == 0
                        and bfs_map[ni, nj] == -1
                    ):
                        bfs_map[ni][nj] = bfs_map[i][j] + 1
                        queue.append((ni, nj))
                        if bfs_map[ni][nj] == adj_steps:
                            adj_cells.append((ni, nj))

            # Sample a goal state from adjacent cells.
            goal_ij = adj_cells[np.random.randint(len(adj_cells))] if len(adj_cells) > 0 else init_ij
        else:
            raise ValueError(f'Unsupported dataset_type: {FLAGS.dataset_type}')

        ob, _ = env.reset(options=dict(task_info=dict(init_ij=init_ij, goal_ij=goal_ij)))

        done = False
        step = 0

        cur_subgoal_dir = None  # Current subgoal direction (only for 'explore').

        while not done:
            if FLAGS.dataset_type == 'explore':
                # Sample a random direction every 10 steps.
                if step % 10 == 0:
                    cur_subgoal_dir = np.random.randn(2)
                    cur_subgoal_dir = cur_subgoal_dir / (np.linalg.norm(cur_subgoal_dir) + 1e-6)
                subgoal_dir = cur_subgoal_dir
            else:
                # Get the oracle subgoal and compute the direction.
                subgoal_xy, _ = env.unwrapped.get_oracle_subgoal(env.unwrapped.get_xy(), env.unwrapped.cur_goal_xy)
                subgoal_dir = subgoal_xy - env.unwrapped.get_xy()
                subgoal_dir = subgoal_dir / (np.linalg.norm(subgoal_dir) + 1e-6)

            agent_ob = env.unwrapped.get_ob(ob_type='states')
            # Exclude the agent's position and add the subgoal direction.
            agent_ob = np.concatenate([agent_ob[2:], subgoal_dir])
            action = actor_fn(agent_ob, temperature=0)
            # Add Gaussian noise to the action.
            action = action + np.random.normal(0, FLAGS.noise, action.shape)
            action = np.clip(action, -1, 1)

            next_ob, reward, terminated, truncated, info = env.step(action)
            done = terminated or truncated
            success = info['success']

            # Sample a new goal state when the current goal is reached.
            if success and FLAGS.dataset_type == 'navigate':
                goal_ij = vertex_cells[np.random.randint(len(vertex_cells))]
                env.unwrapped.set_goal(goal_ij)

            dataset['observations'].append(ob)
            dataset['actions'].append(action)
            dataset['terminals'].append(done)
            dataset['qpos'].append(info['prev_qpos'])
            dataset['qvel'].append(info['prev_qvel'])

            ob = next_ob
            step += 1

        total_steps += step
        if ep_idx < num_train_episodes:
            total_train_steps += step

    print('Total steps:', total_steps)

    train_path = FLAGS.save_path
    val_path = FLAGS.save_path.replace('.npz', '-val.npz')
    pathlib.Path(train_path).parent.mkdir(parents=True, exist_ok=True)

    # Split the dataset into training and validation sets.
    train_dataset = {}
    val_dataset = {}
    for k, v in dataset.items():
        if 'observations' in k and v[0].dtype == np.uint8:
            dtype = np.uint8
        elif k == 'terminals':
            dtype = bool
        else:
            dtype = np.float32
        train_dataset[k] = np.array(v[:total_train_steps], dtype=dtype)
        val_dataset[k] = np.array(v[total_train_steps:], dtype=dtype)

    for path, dataset in [(train_path, train_dataset), (val_path, val_dataset)]:
        np.savez_compressed(path, **dataset)


if __name__ == '__main__':
    app.run(main)



from gymnasium.envs.registration import register

visual_dict = dict(
    ob_type='pixels',
    render_mode='rgb_array',
    width=64,
    height=64,
    camera_name='back',
)
singletask_dict = dict(
    add_noise_to_goal=False,
)

# Environments for offline goal-conditioned RL.
register(
    id='pointmaze-medium-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='point',
        maze_env_type='maze',
        maze_type='medium',
    ),
)
register(
    id='pointmaze-large-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='point',
        maze_env_type='maze',
        maze_type='large',
    ),
)
register(
    id='pointmaze-giant-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='point',
        maze_env_type='maze',
        maze_type='giant',
    ),
)
register(
    id='pointmaze-teleport-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='point',
        maze_env_type='maze',
        maze_type='teleport',
    ),
)

register(
    id='antmaze-medium-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='ant',
        maze_env_type='maze',
        maze_type='medium',
    ),
)
register(
    id='visual-antmaze-medium-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='ant',
        maze_env_type='maze',
        maze_type='medium',
        **visual_dict,
    ),
)
register(
    id='antmaze-large-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='ant',
        maze_env_type='maze',
        maze_type='large',
    ),
)
register(
    id='visual-antmaze-large-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='ant',
        maze_env_type='maze',
        maze_type='large',
        **visual_dict,
    ),
)
register(
    id='antmaze-giant-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='ant',
        maze_env_type='maze',
        maze_type='giant',
    ),
)
register(
    id='visual-antmaze-giant-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='ant',
        maze_env_type='maze',
        maze_type='giant',
        **visual_dict,
    ),
)
register(
    id='antmaze-teleport-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='ant',
        maze_env_type='maze',
        maze_type='teleport',
    ),
)
register(
    id='visual-antmaze-teleport-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='ant',
        maze_env_type='maze',
        maze_type='teleport',
        **visual_dict,
    ),
)

register(
    id='humanoidmaze-medium-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=2000,
    kwargs=dict(
        loco_env_type='humanoid',
        maze_env_type='maze',
        maze_type='medium',
    ),
)
register(
    id='visual-humanoidmaze-medium-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=2000,
    kwargs=dict(
        loco_env_type='humanoid',
        maze_env_type='maze',
        maze_type='medium',
        **visual_dict,
    ),
)
register(
    id='humanoidmaze-large-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=2000,
    kwargs=dict(
        loco_env_type='humanoid',
        maze_env_type='maze',
        maze_type='large',
    ),
)
register(
    id='visual-humanoidmaze-large-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=2000,
    kwargs=dict(
        loco_env_type='humanoid',
        maze_env_type='maze',
        maze_type='large',
        **visual_dict,
    ),
)
register(
    id='humanoidmaze-giant-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=4000,
    kwargs=dict(
        loco_env_type='humanoid',
        maze_env_type='maze',
        maze_type='giant',
    ),
)
register(
    id='visual-humanoidmaze-giant-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=4000,
    kwargs=dict(
        loco_env_type='humanoid',
        maze_env_type='maze',
        maze_type='giant',
        **visual_dict,
    ),
)
register(
    id='humanoidmaze-teleport-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=2000,
    kwargs=dict(
        loco_env_type='humanoid',
        maze_env_type='maze',
        maze_type='teleport',
    ),
)
register(
    id='visual-humanoidmaze-teleport-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=2000,
    kwargs=dict(
        loco_env_type='humanoid',
        maze_env_type='maze',
        maze_type='teleport',
        **visual_dict,
    ),
)

register(
    id='antsoccer-arena-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='ant',
        maze_env_type='ball',
        maze_type='arena',
    ),
)
register(
    id='antsoccer-medium-v0',
    entry_point='ogbench.locomaze.maze:make_maze_env',
    max_episode_steps=1000,
    kwargs=dict(
        loco_env_type='ant',
        maze_env_type='ball',
        maze_type='medium',
    ),
)

# Environments for reward-based single-task offline RL.
for task_id in [None, 1, 2, 3, 4, 5]:
    task_suffix = '' if task_id is None else f'-task{task_id}'
    reward_task_id = 0 if task_id is None else task_id  # 0 means the default task.

    register(
        id=f'pointmaze-medium-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='point',
            maze_env_type='maze',
            maze_type='medium',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'pointmaze-large-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='point',
            maze_env_type='maze',
            maze_type='large',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'pointmaze-giant-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='point',
            maze_env_type='maze',
            maze_type='giant',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'pointmaze-teleport-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='point',
            maze_env_type='maze',
            maze_type='teleport',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )

    register(
        id=f'antmaze-medium-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='ant',
            maze_env_type='maze',
            maze_type='medium',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'visual-antmaze-medium-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='ant',
            maze_env_type='maze',
            maze_type='medium',
            **visual_dict,
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'antmaze-large-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='ant',
            maze_env_type='maze',
            maze_type='large',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'visual-antmaze-large-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='ant',
            maze_env_type='maze',
            maze_type='large',
            **visual_dict,
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'antmaze-giant-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='ant',
            maze_env_type='maze',
            maze_type='giant',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'visual-antmaze-giant-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='ant',
            maze_env_type='maze',
            maze_type='giant',
            **visual_dict,
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'antmaze-teleport-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='ant',
            maze_env_type='maze',
            maze_type='teleport',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'visual-antmaze-teleport-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='ant',
            maze_env_type='maze',
            maze_type='teleport',
            **visual_dict,
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )

    register(
        id=f'humanoidmaze-medium-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=2000,
        kwargs=dict(
            loco_env_type='humanoid',
            maze_env_type='maze',
            maze_type='medium',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'visual-humanoidmaze-medium-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=2000,
        kwargs=dict(
            loco_env_type='humanoid',
            maze_env_type='maze',
            maze_type='medium',
            **visual_dict,
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'humanoidmaze-large-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=2000,
        kwargs=dict(
            loco_env_type='humanoid',
            maze_env_type='maze',
            maze_type='large',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'visual-humanoidmaze-large-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=2000,
        kwargs=dict(
            loco_env_type='humanoid',
            maze_env_type='maze',
            maze_type='large',
            **visual_dict,
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'humanoidmaze-giant-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=4000,
        kwargs=dict(
            loco_env_type='humanoid',
            maze_env_type='maze',
            maze_type='giant',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'visual-humanoidmaze-giant-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=4000,
        kwargs=dict(
            loco_env_type='humanoid',
            maze_env_type='maze',
            maze_type='giant',
            **visual_dict,
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'humanoidmaze-teleport-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=2000,
        kwargs=dict(
            loco_env_type='humanoid',
            maze_env_type='maze',
            maze_type='teleport',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'visual-humanoidmaze-teleport-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=2000,
        kwargs=dict(
            loco_env_type='humanoid',
            maze_env_type='maze',
            maze_type='teleport',
            **visual_dict,
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )

    register(
        id=f'antsoccer-arena-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='ant',
            maze_env_type='ball',
            maze_type='arena',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )
    register(
        id=f'antsoccer-medium-singletask{task_suffix}-v0',
        entry_point='ogbench.locomaze.maze:make_maze_env',
        max_episode_steps=1000,
        kwargs=dict(
            loco_env_type='ant',
            maze_env_type='ball',
            maze_type='medium',
            reward_task_id=reward_task_id,
            **singletask_dict,
        ),
    )



import numpy as np

from ogbench.manipspace.oracles.markov.markov_oracle import MarkovOracle


class CubeMarkovOracle(MarkovOracle):
    def __init__(self, max_step=200, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._max_step = max_step

    def reset(self, ob, info):
        self._done = False
        self._step = 0
        self._max_step = 200
        self._final_pos = np.random.uniform(*self._env.unwrapped._arm_sampling_bounds)
        self._final_yaw = np.random.uniform(-np.pi, np.pi)

    def select_action(self, ob, info):
        effector_pos = info['proprio/effector_pos']
        effector_yaw = info['proprio/effector_yaw'][0]
        gripper_opening = info['proprio/gripper_opening']

        target_block = info['privileged/target_block']
        block_pos = info[f'privileged/block_{target_block}_pos']
        block_yaw = self.shortest_yaw(effector_yaw, info[f'privileged/block_{target_block}_yaw'][0])
        target_pos = info['privileged/target_block_pos']
        target_yaw = self.shortest_yaw(effector_yaw, info['privileged/target_block_yaw'][0])

        block_above_offset = np.array([0, 0, 0.18])
        above_threshold = 0.16
        gripper_closed = info['proprio/gripper_contact'] > 0.5
        gripper_open = info['proprio/gripper_contact'] < 0.1
        above = effector_pos[2] > above_threshold
        xy_aligned = np.linalg.norm(block_pos[:2] - effector_pos[:2]) <= 0.04
        pos_aligned = np.linalg.norm(block_pos - effector_pos) <= 0.02
        target_xy_aligned = np.linalg.norm(target_pos[:2] - block_pos[:2]) <= 0.04
        target_pos_aligned = np.linalg.norm(target_pos - block_pos) <= 0.02
        final_pos_aligned = np.linalg.norm(self._final_pos - effector_pos) <= 0.04

        gain_pos = 5
        gain_yaw = 3
        action = np.zeros(5)
        if not target_pos_aligned:
            if not xy_aligned:
                self.print_phase('1: Move above the block')
                action = np.zeros(5)
                diff = block_pos + block_above_offset - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (block_yaw - effector_yaw) * gain_yaw
                action[4] = -1
            elif not pos_aligned:
                self.print_phase('2: Move to the block')
                diff = block_pos - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (block_yaw - effector_yaw) * gain_yaw
                action[4] = -1
            elif pos_aligned and not gripper_closed:
                self.print_phase('3: Grasp')
                diff = block_pos - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (block_yaw - effector_yaw) * gain_yaw
                action[4] = 1
            elif pos_aligned and gripper_closed and not above and not target_xy_aligned:
                self.print_phase('4: Move in the air')
                diff = np.array([block_pos[0], block_pos[1], block_above_offset[2] * 2]) - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (target_yaw - block_yaw) * gain_yaw
                action[4] = 1
            elif pos_aligned and gripper_closed and above and not target_xy_aligned:
                self.print_phase('5: Move above the target')
                diff = target_pos + block_above_offset - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (target_yaw - block_yaw) * gain_yaw
                action[4] = 1
            else:
                self.print_phase('6: Move to the target')
                diff = target_pos - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (target_yaw - block_yaw) * gain_yaw
                action[4] = 1
        else:
            if not gripper_open:
                self.print_phase('7: Release')
                diff = target_pos - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (target_yaw - block_yaw) * gain_yaw
                action[4] = -1
            elif gripper_open and not above:
                self.print_phase('8: Move in the air')
                diff = np.array([block_pos[0], block_pos[1], above_threshold * 2]) - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (self._final_yaw - effector_yaw) * gain_yaw
                action[4] = -1
            else:
                self.print_phase('9: Move to the final position')
                diff = self._final_pos - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (self._final_yaw - effector_yaw) * gain_yaw
                action[4] = -1

            if final_pos_aligned:
                self._done = True

        action = np.clip(action, -1, 1)
        if self._debug:
            print(action)

        self._step += 1
        if self._step == self._max_step:
            self._done = True

        return action



import json
import os
import random
import time
from collections import defaultdict

import jax
import numpy as np
import tqdm
import wandb
from absl import app, flags
from agents import agents
from ml_collections import config_flags
from utils.datasets import Dataset, GCDataset, HGCDataset
from utils.env_utils import make_env_and_datasets
from utils.evaluation import evaluate
from utils.flax_utils import restore_agent, save_agent
from utils.log_utils import CsvLogger, get_exp_name, get_flag_dict, get_wandb_video, setup_wandb

FLAGS = flags.FLAGS

flags.DEFINE_string('run_group', 'Debug', 'Run group.')
flags.DEFINE_integer('seed', 0, 'Random seed.')
flags.DEFINE_string('env_name', 'antmaze-large-navigate-v0', 'Environment (dataset) name.')
flags.DEFINE_string('save_dir', 'exp/', 'Save directory.')
flags.DEFINE_string('restore_path', None, 'Restore path.')
flags.DEFINE_integer('restore_epoch', None, 'Restore epoch.')

flags.DEFINE_integer('train_steps', 1000000, 'Number of training steps.')
flags.DEFINE_integer('log_interval', 5000, 'Logging interval.')
flags.DEFINE_integer('eval_interval', 100000, 'Evaluation interval.')
flags.DEFINE_integer('save_interval', 1000000, 'Saving interval.')

flags.DEFINE_integer('eval_tasks', None, 'Number of tasks to evaluate (None for all).')
flags.DEFINE_integer('eval_episodes', 20, 'Number of episodes for each task.')
flags.DEFINE_float('eval_temperature', 0, 'Actor temperature for evaluation.')
flags.DEFINE_float('eval_gaussian', None, 'Action Gaussian noise for evaluation.')
flags.DEFINE_integer('video_episodes', 1, 'Number of video episodes for each task.')
flags.DEFINE_integer('video_frame_skip', 3, 'Frame skip for videos.')
flags.DEFINE_integer('eval_on_cpu', 1, 'Whether to evaluate on CPU.')

config_flags.DEFINE_config_file('agent', 'agents/gciql.py', lock_config=False)


def main(_):
    # Set up logger.
    exp_name = get_exp_name(FLAGS.seed)
    setup_wandb(project='OGBench', group=FLAGS.run_group, name=exp_name)

    FLAGS.save_dir = os.path.join(FLAGS.save_dir, wandb.run.project, FLAGS.run_group, exp_name)
    os.makedirs(FLAGS.save_dir, exist_ok=True)
    flag_dict = get_flag_dict()
    with open(os.path.join(FLAGS.save_dir, 'flags.json'), 'w') as f:
        json.dump(flag_dict, f)

    # Set up environment and dataset.
    config = FLAGS.agent
    env, train_dataset, val_dataset = make_env_and_datasets(FLAGS.env_name, frame_stack=config['frame_stack'])

    dataset_class = {
        'GCDataset': GCDataset,
        'HGCDataset': HGCDataset,
    }[config['dataset_class']]
    train_dataset = dataset_class(Dataset.create(**train_dataset), config)
    if val_dataset is not None:
        val_dataset = dataset_class(Dataset.create(**val_dataset), config)

    # Initialize agent.
    random.seed(FLAGS.seed)
    np.random.seed(FLAGS.seed)

    example_batch = train_dataset.sample(1)
    if config['discrete']:
        # Fill with the maximum action to let the agent know the action space size.
        example_batch['actions'] = np.full_like(example_batch['actions'], env.action_space.n - 1)

    agent_class = agents[config['agent_name']]
    agent = agent_class.create(
        FLAGS.seed,
        example_batch['observations'],
        example_batch['actions'],
        config,
    )

    # Restore agent.
    if FLAGS.restore_path is not None:
        agent = restore_agent(agent, FLAGS.restore_path, FLAGS.restore_epoch)

    # Train agent.
    train_logger = CsvLogger(os.path.join(FLAGS.save_dir, 'train.csv'))
    eval_logger = CsvLogger(os.path.join(FLAGS.save_dir, 'eval.csv'))
    first_time = time.time()
    last_time = time.time()
    for i in tqdm.tqdm(range(1, FLAGS.train_steps + 1), smoothing=0.1, dynamic_ncols=True):
        # Update agent.
        batch = train_dataset.sample(config['batch_size'])
        agent, update_info = agent.update(batch)

        # Log metrics.
        if i % FLAGS.log_interval == 0:
            train_metrics = {f'training/{k}': v for k, v in update_info.items()}
            if val_dataset is not None:
                val_batch = val_dataset.sample(config['batch_size'])
                _, val_info = agent.total_loss(val_batch, grad_params=None)
                train_metrics.update({f'validation/{k}': v for k, v in val_info.items()})
            train_metrics['time/epoch_time'] = (time.time() - last_time) / FLAGS.log_interval
            train_metrics['time/total_time'] = time.time() - first_time
            last_time = time.time()
            wandb.log(train_metrics, step=i)
            train_logger.log(train_metrics, step=i)

        # Evaluate agent.
        if i == 1 or i % FLAGS.eval_interval == 0:
            if FLAGS.eval_on_cpu:
                eval_agent = jax.device_put(agent, device=jax.devices('cpu')[0])
            else:
                eval_agent = agent
            renders = []
            eval_metrics = {}
            overall_metrics = defaultdict(list)
            task_infos = env.unwrapped.task_infos if hasattr(env.unwrapped, 'task_infos') else env.task_infos
            num_tasks = FLAGS.eval_tasks if FLAGS.eval_tasks is not None else len(task_infos)
            for task_id in tqdm.trange(1, num_tasks + 1):
                task_name = task_infos[task_id - 1]['task_name']
                eval_info, trajs, cur_renders = evaluate(
                    agent=eval_agent,
                    env=env,
                    task_id=task_id,
                    config=config,
                    num_eval_episodes=FLAGS.eval_episodes,
                    num_video_episodes=FLAGS.video_episodes,
                    video_frame_skip=FLAGS.video_frame_skip,
                    eval_temperature=FLAGS.eval_temperature,
                    eval_gaussian=FLAGS.eval_gaussian,
                )
                renders.extend(cur_renders)
                metric_names = ['success']
                eval_metrics.update(
                    {f'evaluation/{task_name}_{k}': v for k, v in eval_info.items() if k in metric_names}
                )
                for k, v in eval_info.items():
                    if k in metric_names:
                        overall_metrics[k].append(v)
            for k, v in overall_metrics.items():
                eval_metrics[f'evaluation/overall_{k}'] = np.mean(v)

            if FLAGS.video_episodes > 0:
                video = get_wandb_video(renders=renders, n_cols=num_tasks)
                eval_metrics['video'] = video

            wandb.log(eval_metrics, step=i)
            eval_logger.log(eval_metrics, step=i)

        # Save agent.
        if i % FLAGS.save_interval == 0:
            save_agent(agent, FLAGS.save_dir, i)

    train_logger.close()
    eval_logger.close()


if __name__ == '__main__':
    app.run(main)



import mujoco
import numpy as np
from dm_control import mjcf

from ogbench.manipspace import lie
from ogbench.manipspace.envs.manipspace_env import ManipSpaceEnv


class SceneEnv(ManipSpaceEnv):
    """Scene environment.

    This environment consists of a cube, two buttons, a drawer, and a window. The goal is to manipulate the objects
    to a target configuration. The buttons toggle the lock state of the drawer and window.

    In addition to `qpos` and `qvel`, it maintains the following state variables.
    - `button_states`: A binary array of size `num_buttons` representing the state of each button. Stored in
        `_cur_button_states`.
    """

    def __init__(self, env_type, permute_blocks=True, *args, **kwargs):
        """Initialize the Scene environment.

        Args:
            env_type: Unused; defined for compatibility with the other environments.
            permute_blocks: Whether to randomly permute the order of the blocks at task initialization.
            *args: Additional arguments to pass to the parent class.
            **kwargs: Additional keyword arguments to pass to the parent class.
        """
        self._env_type = env_type
        self._permute_blocks = permute_blocks

        super().__init__(*args, **kwargs)

        # Adjust workspace bounds to a smaller region.
        self._arm_sampling_bounds = np.asarray([[0.25, -0.2, 0.20], [0.6, 0.2, 0.35]])
        self._object_sampling_bounds = np.asarray([[0.3, -0.07], [0.45, 0.18]])
        self._target_sampling_bounds = self._object_sampling_bounds

        # Define constants.
        self._drawer_center = np.array([0.33, -0.24, 0.066])
        self._cube_colors = np.array([self._colors['red'], self._colors['blue']])
        self._cube_success_colors = np.array([self._colors['lightred'], self._colors['lightblue']])
        self._num_cubes = 1
        self._num_buttons = 2
        self._num_button_states = 2
        self._cur_button_states = np.array([0] * self._num_buttons)

        # Target info.
        self._target_task = 'cube'
        # The target cube position is stored in the mocap object.
        self._target_block = 0
        self._target_button = 0
        self._target_button_states = np.array([0] * self._num_buttons)
        self._target_drawer_pos = 0.0
        self._target_window_pos = 0.0

    def set_state(self, qpos, qvel, button_states):
        self._cur_button_states = button_states.copy()
        self._apply_button_states()
        super().set_state(qpos, qvel)

    def set_tasks(self):
        self.task_infos = [
            dict(
                task_name='task1_open',
                init=dict(
                    block_xyzs=np.array([[0.35, 0.05, 0.02]]),
                    button_states=np.array([1, 1]),
                    drawer_pos=0.0,
                    window_pos=0.0,
                ),
                goal=dict(
                    block_xyzs=np.array([[0.35, 0.05, 0.02]]),
                    button_states=np.array([1, 1]),
                    drawer_pos=-0.16,
                    window_pos=0.2,
                ),
            ),
            dict(
                task_name='task2_unlock_and_lock',
                init=dict(
                    block_xyzs=np.array([[0.35, -0.05, 0.02]]),
                    button_states=np.array([0, 0]),
                    drawer_pos=-0.16,
                    window_pos=0.2,
                ),
                goal=dict(
                    block_xyzs=np.array([[0.35, -0.05, 0.02]]),
                    button_states=np.array([0, 0]),
                    drawer_pos=0.0,
                    window_pos=0.0,
                ),
            ),
            dict(
                task_name='task3_rearrange_medium',
                init=dict(
                    block_xyzs=np.array([[0.4, -0.05, 0.02]]),
                    button_states=np.array([1, 0]),
                    drawer_pos=0.0,
                    window_pos=0.2,
                ),
                goal=dict(
                    block_xyzs=np.array([[0.4, 0.15, 0.02]]),
                    button_states=np.array([1, 1]),
                    drawer_pos=-0.16,
                    window_pos=0.0,
                ),
            ),
            dict(
                task_name='task4_put_in_drawer',
                init=dict(
                    block_xyzs=np.array([[0.35, 0.05, 0.02]]),
                    button_states=np.array([0, 0]),
                    drawer_pos=0.0,
                    window_pos=0.0,
                ),
                goal=dict(
                    block_xyzs=np.array([[0.33, -0.356, 0.065986]]),
                    button_states=np.array([1, 0]),
                    drawer_pos=0.0,
                    window_pos=0.0,
                ),
            ),
            dict(
                task_name='task5_rearrange_hard',
                init=dict(
                    block_xyzs=np.array([[0.35, 0.15, 0.02]]),
                    button_states=np.array([0, 0]),
                    drawer_pos=0.0,
                    window_pos=0.0,
                ),
                goal=dict(
                    block_xyzs=np.array([[0.33, -0.356, 0.065986]]),
                    button_states=np.array([0, 0]),
                    drawer_pos=0.0,
                    window_pos=0.2,
                ),
            ),
        ]

        if self._reward_task_id == 0:
            self._reward_task_id = 2  # Default task.

    def add_objects(self, arena_mjcf):
        # Add objects to scene.
        cube_mjcf = mjcf.from_path((self._desc_dir / 'cube.xml').as_posix())
        arena_mjcf.include_copy(cube_mjcf)
        button_mjcf = mjcf.from_path((self._desc_dir / 'buttons.xml').as_posix())
        arena_mjcf.include_copy(button_mjcf)
        drawer_mjcf = mjcf.from_path((self._desc_dir / 'drawer.xml').as_posix())
        arena_mjcf.include_copy(drawer_mjcf)
        window_mjcf = mjcf.from_path((self._desc_dir / 'window.xml').as_posix())
        arena_mjcf.include_copy(window_mjcf)

        # Save geoms.
        self._cube_geoms_list = []
        for i in range(self._num_cubes):
            self._cube_geoms_list.append(cube_mjcf.find('body', f'object_{i}').find_all('geom'))
        self._cube_target_geoms_list = []
        for i in range(self._num_cubes):
            self._cube_target_geoms_list.append(cube_mjcf.find('body', f'object_target_{i}').find_all('geom'))
        self._button_geoms_list = []
        for i in range(self._num_buttons):
            self._button_geoms_list.append([button_mjcf.find('geom', f'btngeom_{i}')])

        # Add cameras.
        cameras = {
            'front': {
                'pos': (1.139, 0.000, 0.821),
                'xyaxes': (0.000, 1.000, 0.000, -0.627, 0.000, 0.779),
            },
            'front_pixels': {
                'pos': (0.905, 0.000, 0.762),
                'xyaxes': (0.000, 1.000, 0.000, -0.771, 0.000, 0.637),
            },
        }
        for camera_name, camera_kwargs in cameras.items():
            arena_mjcf.worldbody.add('camera', name=camera_name, **camera_kwargs)

    def post_compilation_objects(self):
        # Cube geom IDs.
        self._cube_geom_ids_list = [
            [self._model.geom(geom.full_identifier).id for geom in cube_geoms] for cube_geoms in self._cube_geoms_list
        ]
        self._cube_target_mocap_ids = [
            self._model.body(f'object_target_{i}').mocapid[0] for i in range(self._num_cubes)
        ]
        self._cube_target_geom_ids_list = [
            [self._model.geom(geom.full_identifier).id for geom in cube_target_geoms]
            for cube_target_geoms in self._cube_target_geoms_list
        ]

        # Button geom IDs.
        self._button_geom_ids_list = [
            [self._model.geom(geom.full_identifier).id for geom in button_geoms]
            for button_geoms in self._button_geoms_list
        ]
        self._button_site_ids = [self._model.site(f'btntop_{i}').id for i in range(self._num_buttons)]

        # Drawer and window site IDs.
        self._drawer_site_id = self._model.site('drawer_handle_center').id
        self._drawer_target_site_id = self._model.site('drawer_handle_center_target').id

        self._window_site_id = self._model.site('window_handle_center').id
        self._window_target_site_id = self._model.site('window_handle_center_target').id

    def _apply_button_states(self):
        # Adjust button colors based on the current state.
        for i in range(self._num_buttons):
            for gid in self._button_geom_ids_list[i]:
                self._model.geom(gid).rgba = self._colors['red' if self._cur_button_states[i] == 0 else 'white']

        # Lock or unlock the drawer and window based on the button states.
        # We adjust the damping of the joints to lock the drawer and window. This needs to be set carefully because
        # setting it to a very high value can cause numerical instability. We use 1e6. This is a reasonably safe value,
        # but it still allows the drawer and window to move very slightly with a strong enough force. We also tested
        # 1e7, but it caused numerical instability when interacting with the cube.
        if self._cur_button_states[0] == 0:
            # Set the damping to a high value to lock the drawer.
            self._model.joint('drawer_slide').damping[0] = 1e6
            self._model.material('drawer_handle').rgba = self._colors['red']
        else:
            # Unset the damping to unlock the drawer.
            self._model.joint('drawer_slide').damping[0] = 2.0
            self._model.material('drawer_handle').rgba = self._colors['white']
        if self._cur_button_states[1] == 0:
            # Set the damping to a high value to lock the window.
            self._model.joint('window_slide').damping[0] = 1e6
            self._model.material('window_handle').rgba = self._colors['red']
        else:
            # Unset the damping to unlock the window.
            self._model.joint('window_slide').damping[0] = 2.0
            self._model.material('window_handle').rgba = self._colors['white']

        mujoco.mj_forward(self._model, self._data)

    def initialize_episode(self):
        # Set cube colors.
        for i in range(self._num_cubes):
            for gid in self._cube_geom_ids_list[i]:
                self._model.geom(gid).rgba = self._cube_colors[i]
            for gid in self._cube_target_geom_ids_list[i]:
                self._model.geom(gid).rgba[:3] = self._cube_colors[i, :3]

        self._data.qpos[self._arm_joint_ids] = self._home_qpos
        mujoco.mj_kinematics(self._model, self._data)

        if self._mode == 'data_collection':
            # Randomize the scene.

            self.initialize_arm()

            # Randomize block positions and orientations.
            for i in range(self._num_cubes):
                xy = self.np_random.uniform(*self._object_sampling_bounds)
                obj_pos = (*xy, 0.02)
                yaw = self.np_random.uniform(0, 2 * np.pi)
                obj_ori = lie.SO3.from_z_radians(yaw).wxyz.tolist()
                self._data.joint(f'object_joint_{i}').qpos[:3] = obj_pos
                self._data.joint(f'object_joint_{i}').qpos[3:] = obj_ori

            # Randomize button states.
            for i in range(self._num_buttons):
                self._cur_button_states[i] = self.np_random.choice(self._num_button_states)
            self._apply_button_states()

            # Randomize drawer and window positions.
            self._data.joint('drawer_slide').qpos[0] = self.np_random.uniform(-0.16, 0)
            self._data.joint('window_slide').qpos[0] = self.np_random.uniform(0, 0.2)

            # Set a new target.
            self.set_new_target(return_info=False)
        else:
            # Set object positions and orientations based on the current task.

            if self._permute_blocks:
                # Randomize the order of the cubes when there are multiple cubes.
                block_permutation = self.np_random.permutation(self._num_cubes)
            else:
                block_permutation = np.arange(self._num_cubes)
            init_block_xyzs = self.cur_task_info['init']['block_xyzs'].copy()[block_permutation]
            goal_block_xyzs = self.cur_task_info['goal']['block_xyzs'].copy()[block_permutation]
            # Get the current task info for the other objects.
            init_button_states = self.cur_task_info['init']['button_states'].copy()
            goal_button_states = self.cur_task_info['goal']['button_states'].copy()
            init_drawer_pos = self.cur_task_info['init']['drawer_pos']
            goal_drawer_pos = self.cur_task_info['goal']['drawer_pos']
            init_window_pos = self.cur_task_info['init']['window_pos']
            goal_window_pos = self.cur_task_info['goal']['window_pos']

            # First, force set the current scene to the goal state to obtain the goal observation.
            saved_qpos = self._data.qpos.copy()
            saved_qvel = self._data.qvel.copy()
            self.initialize_arm()
            for i in range(self._num_cubes):
                self._data.joint(f'object_joint_{i}').qpos[:3] = goal_block_xyzs[i]
                self._data.joint(f'object_joint_{i}').qpos[3:] = lie.SO3.identity().wxyz.tolist()
                self._data.mocap_pos[self._cube_target_mocap_ids[i]] = goal_block_xyzs[i]
                self._data.mocap_quat[self._cube_target_mocap_ids[i]] = lie.SO3.identity().wxyz.tolist()
            self._cur_button_states = goal_button_states.copy()
            self._apply_button_states()
            self._data.joint('drawer_slide').qpos[0] = goal_drawer_pos
            self._data.joint('window_slide').qpos[0] = goal_window_pos
            mujoco.mj_forward(self._model, self._data)

            # Do a few random steps to make the scene stable.
            for _ in range(2):
                self.step(self.action_space.sample())

            # Save the goal observation.
            self._cur_goal_ob = (
                self.compute_oracle_observation() if self._use_oracle_rep else self.compute_observation()
            )
            if self._render_goal:
                self._cur_goal_rendered = self.render()
            else:
                self._cur_goal_rendered = None

            # Now, do the actual reset.
            self._data.qpos[:] = saved_qpos
            self._data.qvel[:] = saved_qvel
            self.initialize_arm()
            for i in range(self._num_cubes):
                # Randomize the position and orientation of the cube slightly.
                obj_pos = init_block_xyzs[i].copy()
                obj_pos[:2] += self.np_random.uniform(-0.01, 0.01, size=2)
                self._data.joint(f'object_joint_{i}').qpos[:3] = obj_pos
                yaw = self.np_random.uniform(0, 2 * np.pi)
                obj_ori = lie.SO3.from_z_radians(yaw).wxyz.tolist()
                self._data.joint(f'object_joint_{i}').qpos[3:] = obj_ori
                self._data.mocap_pos[self._cube_target_mocap_ids[i]] = goal_block_xyzs[i]
                self._data.mocap_quat[self._cube_target_mocap_ids[i]] = lie.SO3.identity().wxyz.tolist()
            # Set the button states.
            self._cur_button_states = init_button_states.copy()
            self._target_button_states = goal_button_states.copy()
            self._apply_button_states()
            # Randomize the drawer and window positions slightly.
            self._data.joint('drawer_slide').qpos[0] = np.clip(
                init_drawer_pos + self.np_random.uniform(-0.01, 0.01), -0.16, 0
            )
            self._model.site('drawer_handle_center_target').pos[1] = goal_drawer_pos
            self._target_drawer_pos = goal_drawer_pos
            self._data.joint('window_slide').qpos[0] = np.clip(
                init_window_pos + self.np_random.uniform(-0.01, 0.01), 0, 0.2
            )
            self._model.site('window_handle_center_target').pos[0] = goal_window_pos
            self._target_window_pos = goal_window_pos

        # Forward kinematics to update site positions.
        self.pre_step()
        mujoco.mj_forward(self._model, self._data)
        self.post_step()

        self._success = False

    def _is_in_drawer(self, obj_pos):
        """Check if the object is in the drawer."""
        drawer_pos_y = self._data.site_xpos[self._drawer_site_id][1]
        drawer_low = np.array([0.21, drawer_pos_y - 0.27, 0.0])
        drawer_high = np.array([0.45, drawer_pos_y - 0.07, 0.15])
        return np.all(drawer_low <= obj_pos) and np.all(obj_pos <= drawer_high)

    def set_new_target(self, return_info=True, p_stack=0.5):
        """Set a new random target for data collection.

        Args:
            return_info: Whether to return the observation and reset info.
            p_stack: Probability of stacking the target block on top of another block when there are multiple blocks
                and the target task is 'cube'.
        """
        assert self._mode == 'data_collection'

        # Only consider blocks not in the drawer.
        available_blocks = []
        for i in range(self._num_cubes):
            if not self._is_in_drawer(self._data.joint(f'object_joint_{i}').qpos[:3]):
                available_blocks.append(i)

        # Probability of each task.
        p_cube = 1.0 if len(available_blocks) > 0 else 0.0
        p_button = 1.0
        p_drawer = 0.25 if self._cur_button_states[0] == 0 else 1.0
        p_window = 0.25 if self._cur_button_states[1] == 0 else 1.0
        probs = np.array([p_cube, p_button, p_drawer, p_window])
        probs /= probs.sum()

        # Probability of putting the target block in the drawer when the target task is 'cube'.
        p_put_in_drawer = 0.3

        self._target_task = self.np_random.choice(['cube', 'button', 'drawer', 'window'], p=probs)

        if self._target_task == 'cube':
            # Set cube target.
            block_xyzs = np.array([self._data.joint(f'object_joint_{i}').qpos[:3] for i in range(self._num_cubes)])

            # Compute the top blocks.
            top_blocks = []
            for i in range(self._num_cubes):
                if i not in available_blocks:
                    continue
                for j in range(self._num_cubes):
                    if i == j:
                        continue
                    if (
                        block_xyzs[j][2] > block_xyzs[i][2]
                        and np.linalg.norm(block_xyzs[i][:2] - block_xyzs[j][:2]) < 0.02
                    ):
                        break
                else:
                    top_blocks.append(i)

            # Pick one of the top cubes as the target.
            self._target_block = self.np_random.choice(top_blocks)

            put_in_drawer = (
                self._data.joint('drawer_slide').qpos[0] < -0.12 and self.np_random.uniform() < p_put_in_drawer
            )
            stack = len(top_blocks) >= 2 and self.np_random.uniform() < p_stack
            if put_in_drawer:
                # Put the target block in the drawer.
                tar_pos = self._drawer_center.copy()
                tar_pos[:2] = tar_pos[:2] + self.np_random.uniform(-0.005, 0.005, size=2)
            elif stack:
                # Stack the target block on top of another block.
                block_idx = self.np_random.choice(list(set(top_blocks) - {self._target_block}))
                block_pos = self._data.joint(f'object_joint_{block_idx}').qpos[:3]
                tar_pos = np.array([block_pos[0], block_pos[1], block_pos[2] + 0.04])
            else:
                # Randomize target position.
                xy = self.np_random.uniform(*self._target_sampling_bounds)
                tar_pos = (*xy, 0.02)
            # Randomize target orientation.
            yaw = self.np_random.uniform(0, 2 * np.pi)
            tar_ori = lie.SO3.from_z_radians(yaw).wxyz.tolist()

            # Only show the target block.
            for i in range(self._num_cubes):
                if i == self._target_block:
                    # Set the target position and orientation.
                    self._data.mocap_pos[self._cube_target_mocap_ids[i]] = tar_pos
                    self._data.mocap_quat[self._cube_target_mocap_ids[i]] = tar_ori
                else:
                    # Move the non-target blocks out of the way.
                    self._data.mocap_pos[self._cube_target_mocap_ids[i]] = (0, 0, -0.3)
                    self._data.mocap_quat[self._cube_target_mocap_ids[i]] = lie.SO3.identity().wxyz.tolist()

            # Set the target colors.
            for i in range(self._num_cubes):
                if self._visualize_info and i == self._target_block:
                    for gid in self._cube_target_geom_ids_list[i]:
                        self._model.geom(gid).rgba[3] = 0.2
                else:
                    for gid in self._cube_target_geom_ids_list[i]:
                        self._model.geom(gid).rgba[3] = 0.0
        elif self._target_task == 'button':
            # Set target button.
            self._target_button = self.np_random.choice(self._num_buttons)
            self._target_button_states[self._target_button] = (
                self._cur_button_states[self._target_button] + 1
            ) % self._num_button_states
        elif self._target_task == 'drawer':
            # Set target drawer position.
            if self._data.joint('drawer_slide').qpos[0] >= -0.08:  # Drawer closed.
                self._target_drawer_pos = -0.16
            else:  # Drawer open.
                self._target_drawer_pos = 0.0
            self._model.site('drawer_handle_center_target').pos[1] = self._target_drawer_pos
        elif self._target_task == 'window':
            # Set target window position.
            if self._data.joint('window_slide').qpos[0] <= 0.1:  # Window closed.
                self._target_window_pos = 0.2
            else:  # Window open.
                self._target_window_pos = 0.0
            self._model.site('window_handle_center_target').pos[0] = self._target_window_pos

        mujoco.mj_kinematics(self._model, self._data)

        if return_info:
            return self.compute_observation(), self.get_reset_info()

    def pre_step(self):
        self._prev_button_states = self._cur_button_states.copy()
        super().pre_step()

    def _compute_successes(self):
        """Compute object successes."""
        cube_successes = []
        for i in range(self._num_cubes):
            obj_pos = self._data.joint(f'object_joint_{i}').qpos[:3]
            tar_pos = self._data.mocap_pos[self._cube_target_mocap_ids[i]]
            if np.linalg.norm(obj_pos - tar_pos) <= 0.04:
                cube_successes.append(True)
            else:
                cube_successes.append(False)
        button_successes = [
            (self._cur_button_states[i] == self._target_button_states[i]) for i in range(self._num_buttons)
        ]
        drawer_success = np.abs(self._data.joint('drawer_slide').qpos[0] - self._target_drawer_pos) <= 0.04
        window_success = np.abs(self._data.joint('window_slide').qpos[0] - self._target_window_pos) <= 0.04

        return cube_successes, button_successes, drawer_success, window_success

    def post_step(self):
        # Check numerical stability.
        if self._mode == 'task':
            # Very rarely, the blocks can go out of bounds due to numerical instability. This can (rarely) happen
            # when the robot presses the drawer lock button while the drawer is moving and the block is in the drawer.
            # We only check this in the task mode, because we will manually filter out these cases outside the class in
            # the data collection mode with a more stringent check.
            is_healthy = True
            for i in range(self._num_cubes):
                obj_pos = self._data.joint(f'object_joint_{i}').qpos[:3]
                # Check if the object is out of bounds.
                if np.any(obj_pos <= self._workspace_bounds[0] - 0.2) or np.any(
                    obj_pos >= self._workspace_bounds[1] + 0.2
                ):
                    is_healthy = False
                    break

            if not is_healthy:
                # Manually reset the cube position to a random initial position.
                print('Numerical instability detected. Resetting cube positions.', flush=True)
                for i in range(self._num_cubes):
                    xy = self.np_random.uniform(*self._object_sampling_bounds)
                    obj_pos = (*xy, 0.02)
                    yaw = self.np_random.uniform(0, 2 * np.pi)
                    obj_ori = lie.SO3.from_z_radians(yaw).wxyz.tolist()
                    self._data.joint(f'object_joint_{i}').qpos[:3] = obj_pos
                    self._data.joint(f'object_joint_{i}').qpos[3:] = obj_ori
                    self._data.joint('object_joint_0').qvel[:] = 0.0
                mujoco.mj_forward(self._model, self._data)

        # Update button states.
        for i in range(self._num_buttons):
            prev_joint_pos = self._prev_ob_info[f'privileged/button_{i}_pos'][0]
            cur_joint_pos = self._data.joint(f'buttonbox_joint_{i}').qpos.copy()[0]
            if prev_joint_pos > -0.02 and cur_joint_pos <= -0.02:
                # Button pressed: change the state of the button.
                self._cur_button_states[i] = (self._cur_button_states[i] + 1) % self._num_button_states
        self._apply_button_states()

        # Evaluate successes.
        cube_successes, button_successes, drawer_success, window_success = self._compute_successes()
        if self._mode == 'data_collection':
            self._success = {
                'cube': cube_successes[self._target_block],
                'button': button_successes[self._target_button],
                'drawer': drawer_success,
                'window': window_success,
            }[self._target_task]
        else:
            self._success = all(cube_successes) and all(button_successes) and drawer_success and window_success

        # Adjust the colors of the cubes based on success.
        for i in range(self._num_cubes):
            if self._visualize_info and (self._mode == 'task' or i == self._target_block):
                for gid in self._cube_target_geom_ids_list[i]:
                    self._model.geom(gid).rgba[3] = 0.2
            else:
                for gid in self._cube_target_geom_ids_list[i]:
                    self._model.geom(gid).rgba[3] = 0.0

            if self._visualize_info and cube_successes[i]:
                for gid in self._cube_geom_ids_list[i]:
                    self._model.geom(gid).rgba[:3] = self._cube_success_colors[i, :3]
            else:
                for gid in self._cube_geom_ids_list[i]:
                    self._model.geom(gid).rgba[:3] = self._cube_colors[i, :3]

    def add_object_info(self, ob_info):
        # Cube positions and orientations.
        for i in range(self._num_cubes):
            ob_info[f'privileged/block_{i}_pos'] = self._data.joint(f'object_joint_{i}').qpos[:3].copy()
            ob_info[f'privileged/block_{i}_quat'] = self._data.joint(f'object_joint_{i}').qpos[3:].copy()
            ob_info[f'privileged/block_{i}_yaw'] = np.array(
                [lie.SO3(wxyz=self._data.joint(f'object_joint_{i}').qpos[3:]).compute_yaw_radians()]
            )

        # Button states.
        for i in range(self._num_buttons):
            ob_info[f'privileged/button_{i}_state'] = self._cur_button_states[i]
            ob_info[f'privileged/button_{i}_pos'] = self._data.joint(f'buttonbox_joint_{i}').qpos.copy()
            ob_info[f'privileged/button_{i}_vel'] = self._data.joint(f'buttonbox_joint_{i}').qvel.copy()

        # Drawer states.
        ob_info['privileged/drawer_pos'] = self._data.joint('drawer_slide').qpos.copy()
        ob_info['privileged/drawer_vel'] = self._data.joint('drawer_slide').qvel.copy()
        ob_info['privileged/drawer_handle_pos'] = self._data.site_xpos[self._drawer_site_id].copy()
        ob_info['privileged/drawer_handle_yaw'] = np.array(
            [lie.SO3.from_matrix(self._data.site_xmat[self._drawer_site_id].reshape(3, 3)).compute_yaw_radians()]
        )

        # Window states.
        ob_info['privileged/window_pos'] = self._data.joint('window_slide').qpos.copy()
        ob_info['privileged/window_vel'] = self._data.joint('window_slide').qvel.copy()
        ob_info['privileged/window_handle_pos'] = self._data.site_xpos[self._window_site_id].copy()
        ob_info['privileged/window_handle_yaw'] = np.array(
            [lie.SO3.from_matrix(self._data.site_xmat[self._window_site_id].reshape(3, 3)).compute_yaw_radians()]
        )

        if self._mode == 'data_collection':
            ob_info['privileged/target_task'] = self._target_task

            # Target cube info.
            target_mocap_id = self._cube_target_mocap_ids[self._target_block]
            ob_info['privileged/target_block'] = self._target_block
            ob_info['privileged/target_block_pos'] = self._data.mocap_pos[target_mocap_id].copy()
            ob_info['privileged/target_block_yaw'] = np.array(
                [lie.SO3(wxyz=self._data.mocap_quat[target_mocap_id]).compute_yaw_radians()]
            )

            # Target button info.
            ob_info['privileged/target_button'] = self._target_button
            ob_info['privileged/target_button_state'] = self._target_button_states[self._target_button]
            ob_info['privileged/target_button_top_pos'] = self._data.site_xpos[
                self._button_site_ids[self._target_button]
            ].copy()

            # Target drawer info.
            ob_info['privileged/target_drawer_pos'] = np.array([self._target_drawer_pos])
            ob_info['privileged/target_drawer_handle_pos'] = self._data.site_xpos[self._drawer_target_site_id].copy()

            # Target window info.
            ob_info['privileged/target_window_pos'] = np.array([self._target_window_pos])
            ob_info['privileged/target_window_handle_pos'] = self._data.site_xpos[self._window_target_site_id].copy()

        ob_info['prev_button_states'] = self._prev_button_states.copy()
        ob_info['button_states'] = self._cur_button_states.copy()

    def compute_observation(self):
        if self._ob_type == 'pixels':
            return self.get_pixel_observation()
        else:
            xyz_center = np.array([0.425, 0.0, 0.0])
            xyz_scaler = 10.0
            gripper_scaler = 3.0
            button_scaler = 120.0
            drawer_scaler = 18.0
            window_scaler = 15.0

            ob_info = self.compute_ob_info()
            ob = [
                ob_info['proprio/joint_pos'],
                ob_info['proprio/joint_vel'],
                (ob_info['proprio/effector_pos'] - xyz_center) * xyz_scaler,
                np.cos(ob_info['proprio/effector_yaw']),
                np.sin(ob_info['proprio/effector_yaw']),
                ob_info['proprio/gripper_opening'] * gripper_scaler,
                ob_info['proprio/gripper_contact'],
            ]
            for i in range(self._num_cubes):
                ob.extend(
                    [
                        (ob_info[f'privileged/block_{i}_pos'] - xyz_center) * xyz_scaler,
                        ob_info[f'privileged/block_{i}_quat'],
                        np.cos(ob_info[f'privileged/block_{i}_yaw']),
                        np.sin(ob_info[f'privileged/block_{i}_yaw']),
                    ]
                )
            for i in range(self._num_buttons):
                button_state = np.eye(self._num_button_states)[self._cur_button_states[i]]
                ob.extend(
                    [
                        button_state,
                        ob_info[f'privileged/button_{i}_pos'] * button_scaler,
                        ob_info[f'privileged/button_{i}_vel'],
                    ]
                )
            ob.extend(
                [
                    ob_info['privileged/drawer_pos'] * drawer_scaler,
                    ob_info['privileged/drawer_vel'],
                    ob_info['privileged/window_pos'] * window_scaler,
                    ob_info['privileged/window_vel'],
                ]
            )

            return np.concatenate(ob)

    def compute_oracle_observation(self):
        """Return the oracle goal representation of the current state."""
        xyz_center = np.array([0.425, 0.0, 0.0])
        xyz_scaler = 10.0
        drawer_scaler = 18.0
        window_scaler = 15.0

        ob_info = self.compute_ob_info()
        ob = []
        for i in range(self._num_cubes):
            ob.append((ob_info[f'privileged/block_{i}_pos'] - xyz_center) * xyz_scaler)
        ob.append(self._cur_button_states.astype(np.float64))
        ob.extend(
            [
                ob_info['privileged/drawer_pos'] * drawer_scaler,
                ob_info['privileged/window_pos'] * window_scaler,
            ]
        )

        return np.concatenate(ob)

    def compute_reward(self, ob, action):
        if self._reward_task_id is None:
            return super().compute_reward(ob, action)

        # Compute the reward based on the task.
        cube_successes, button_successes, drawer_success, window_success = self._compute_successes()
        successes = cube_successes + button_successes + [drawer_success, window_success]
        reward = float(sum(successes) - len(successes))
        return reward



from collections import defaultdict

import jax
import numpy as np
from tqdm import trange


def supply_rng(f, rng=jax.random.PRNGKey(0)):
    """Helper function to split the random number generator key before each call to the function."""

    def wrapped(*args, **kwargs):
        nonlocal rng
        rng, key = jax.random.split(rng)
        return f(*args, seed=key, **kwargs)

    return wrapped


def flatten(d, parent_key='', sep='.'):
    """Flatten a dictionary."""
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if hasattr(v, 'items'):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def add_to(dict_of_lists, single_dict):
    """Append values to the corresponding lists in the dictionary."""
    for k, v in single_dict.items():
        dict_of_lists[k].append(v)


def evaluate(
    agent,
    env,
    task_id=None,
    config=None,
    num_eval_episodes=50,
    num_video_episodes=0,
    video_frame_skip=3,
    eval_temperature=0,
    eval_gaussian=None,
):
    """Evaluate the agent in the environment.

    Args:
        agent: Agent.
        env: Environment.
        task_id: Task ID to be passed to the environment.
        config: Configuration dictionary.
        num_eval_episodes: Number of episodes to evaluate the agent.
        num_video_episodes: Number of episodes to render. These episodes are not included in the statistics.
        video_frame_skip: Number of frames to skip between renders.
        eval_temperature: Action sampling temperature.
        eval_gaussian: Standard deviation of the Gaussian noise to add to the actions.

    Returns:
        A tuple containing the statistics, trajectories, and rendered videos.
    """
    actor_fn = supply_rng(agent.sample_actions, rng=jax.random.PRNGKey(np.random.randint(0, 2**32)))
    trajs = []
    stats = defaultdict(list)

    renders = []
    for i in trange(num_eval_episodes + num_video_episodes):
        traj = defaultdict(list)
        should_render = i >= num_eval_episodes

        observation, info = env.reset(options=dict(task_id=task_id, render_goal=should_render))
        goal = info.get('goal')
        goal_frame = info.get('goal_rendered')
        done = False
        step = 0
        render = []
        while not done:
            action = actor_fn(observations=observation, goals=goal, temperature=eval_temperature)
            action = np.array(action)
            if not config.get('discrete'):
                if eval_gaussian is not None:
                    action = np.random.normal(action, eval_gaussian)
                action = np.clip(action, -1, 1)

            next_observation, reward, terminated, truncated, info = env.step(action)
            done = terminated or truncated
            step += 1

            if should_render and (step % video_frame_skip == 0 or done):
                frame = env.render().copy()
                if goal_frame is not None:
                    render.append(np.concatenate([goal_frame, frame], axis=0))
                else:
                    render.append(frame)

            transition = dict(
                observation=observation,
                next_observation=next_observation,
                action=action,
                reward=reward,
                done=done,
                info=info,
            )
            add_to(traj, transition)
            observation = next_observation
        if i < num_eval_episodes:
            add_to(stats, flatten(info))
            trajs.append(traj)
        else:
            renders.append(np.array(render))

    for k, v in stats.items():
        stats[k] = np.mean(v)

    return stats, trajs, renders



from gymnasium.envs.registration import register

register(
    id='powderworld-easy-v0',
    entry_point='ogbench.powderworld.powderworld_env:PowderworldEnv',
    max_episode_steps=500,
    kwargs=dict(num_elems=2),
)

register(
    id='powderworld-medium-v0',
    entry_point='ogbench.powderworld.powderworld_env:PowderworldEnv',
    max_episode_steps=500,
    kwargs=dict(num_elems=5),
)

register(
    id='powderworld-hard-v0',
    entry_point='ogbench.powderworld.powderworld_env:PowderworldEnv',
    max_episode_steps=500,
    kwargs=dict(num_elems=8),
)



import gymnasium
import numpy as np
from gymnasium.spaces import Box


class GymXYWrapper(gymnasium.Wrapper):
    """Wrapper for directional locomotion tasks."""

    def __init__(self, env, resample_interval=100):
        """Initialize the wrapper.

        Args:
            env: Environment.
            resample_interval: Interval at which to resample the direction.
        """
        super().__init__(env)

        self.z = None
        self.num_steps = 0
        self.resample_interval = resample_interval

        ob, _ = self.reset()
        self.observation_space = Box(low=-np.inf, high=np.inf, shape=ob.shape, dtype=np.float64)

    def reset(self, *args, **kwargs):
        ob, info = self.env.reset(*args, **kwargs)
        self.z = np.random.randn(2)
        self.z = self.z / np.linalg.norm(self.z)
        self.num_steps = 0

        return np.concatenate([ob, self.z]), info

    def step(self, action):
        cur_xy = self.unwrapped.data.qpos[:2].copy()
        ob, reward, terminated, truncated, info = self.env.step(action)
        next_xy = self.unwrapped.data.qpos[:2].copy()
        self.num_steps += 1

        # Reward is the dot product of the direction and the change in xy.
        reward = (next_xy - cur_xy).dot(self.z)
        info['xy'] = next_xy
        info['direction'] = self.z

        if self.num_steps % self.resample_interval == 0:
            self.z = np.random.randn(2)
            self.z = self.z / np.linalg.norm(self.z)

        return np.concatenate([ob, self.z]), reward, terminated, truncated, info


class DMCHumanoidXYWrapper(GymXYWrapper):
    """Wrapper for the directional Humanoid task."""

    def step(self, action):
        from ogbench.online_locomotion.humanoid import tolerance

        cur_xy = self.unwrapped.data.qpos[:2].copy()
        ob, reward, terminated, truncated, info = self.env.step(action)
        next_xy = self.unwrapped.data.qpos[:2].copy()
        self.num_steps += 1

        head_height = self.unwrapped.data.xpos[2, 2]  # ['head', 'z']
        torso_upright = self.unwrapped.data.xmat[1, 8]  # ['torso', 'zz']

        standing = tolerance(head_height, bounds=(1.4, float('inf')), margin=1.4 / 4)
        upright = tolerance(torso_upright, bounds=(0.9, float('inf')), margin=1.9, sigmoid='linear', value_at_margin=0)
        stand_reward = standing * upright

        # Reward is the dot product of the direction and the change in xy, multiplied by the stand reward to encourage
        # the agent to stand.
        reward = stand_reward * (1 + (next_xy - cur_xy).dot(self.z) * 100)

        info['xy'] = next_xy
        info['direction'] = self.z

        if self.num_steps % self.resample_interval == 0:
            self.z = np.random.randn(2)
            self.z = self.z / np.linalg.norm(self.z)

        return np.concatenate([ob, self.z]), reward, terminated, truncated, info



import copy
from typing import Any

import flax
import jax
import jax.numpy as jnp
import ml_collections
import optax
from utils.encoders import GCEncoder, encoder_modules
from utils.flax_utils import ModuleDict, TrainState, nonpytree_field
from utils.networks import GCActor, GCDiscreteActor, GCValue


class GCIVLAgent(flax.struct.PyTreeNode):
    """Goal-conditioned implicit V-learning (GCIVL) agent.

    This is a variant of GCIQL that only uses a V function, without Q functions.
    """

    rng: Any
    network: Any
    config: Any = nonpytree_field()

    @staticmethod
    def expectile_loss(adv, diff, expectile):
        """Compute the expectile loss."""
        weight = jnp.where(adv >= 0, expectile, (1 - expectile))
        return weight * (diff**2)

    def value_loss(self, batch, grad_params):
        """Compute the IVL value loss.

        This value loss is similar to the original IQL value loss, but involves additional tricks to stabilize training.
        For example, when computing the expectile loss, we separate the advantage part (which is used to compute the
        weight) and the difference part (which is used to compute the loss), where we use the target value function to
        compute the former and the current value function to compute the latter. This is similar to how double DQN
        mitigates overestimation bias.
        """
        (next_v1_t, next_v2_t) = self.network.select('target_value')(batch['next_observations'], batch['value_goals'])
        next_v_t = jnp.minimum(next_v1_t, next_v2_t)
        q = batch['rewards'] + self.config['discount'] * batch['masks'] * next_v_t

        (v1_t, v2_t) = self.network.select('target_value')(batch['observations'], batch['value_goals'])
        v_t = (v1_t + v2_t) / 2
        adv = q - v_t

        q1 = batch['rewards'] + self.config['discount'] * batch['masks'] * next_v1_t
        q2 = batch['rewards'] + self.config['discount'] * batch['masks'] * next_v2_t
        (v1, v2) = self.network.select('value')(batch['observations'], batch['value_goals'], params=grad_params)
        v = (v1 + v2) / 2

        value_loss1 = self.expectile_loss(adv, q1 - v1, self.config['expectile']).mean()
        value_loss2 = self.expectile_loss(adv, q2 - v2, self.config['expectile']).mean()
        value_loss = value_loss1 + value_loss2

        return value_loss, {
            'value_loss': value_loss,
            'v_mean': v.mean(),
            'v_max': v.max(),
            'v_min': v.min(),
        }

    def actor_loss(self, batch, grad_params, rng=None):
        """Compute the AWR actor loss."""
        v1, v2 = self.network.select('value')(batch['observations'], batch['actor_goals'])
        nv1, nv2 = self.network.select('value')(batch['next_observations'], batch['actor_goals'])
        v = (v1 + v2) / 2
        nv = (nv1 + nv2) / 2
        adv = nv - v

        exp_a = jnp.exp(adv * self.config['alpha'])
        exp_a = jnp.minimum(exp_a, 100.0)

        dist = self.network.select('actor')(batch['observations'], batch['actor_goals'], params=grad_params)
        log_prob = dist.log_prob(batch['actions'])

        actor_loss = -(exp_a * log_prob).mean()

        actor_info = {
            'actor_loss': actor_loss,
            'adv': adv.mean(),
            'bc_log_prob': log_prob.mean(),
        }
        if not self.config['discrete']:
            actor_info.update(
                {
                    'mse': jnp.mean((dist.mode() - batch['actions']) ** 2),
                    'std': jnp.mean(dist.scale_diag),
                }
            )

        return actor_loss, actor_info

    @jax.jit
    def total_loss(self, batch, grad_params, rng=None):
        """Compute the total loss."""
        info = {}
        rng = rng if rng is not None else self.rng

        value_loss, value_info = self.value_loss(batch, grad_params)
        for k, v in value_info.items():
            info[f'value/{k}'] = v

        rng, actor_rng = jax.random.split(rng)
        actor_loss, actor_info = self.actor_loss(batch, grad_params, actor_rng)
        for k, v in actor_info.items():
            info[f'actor/{k}'] = v

        loss = value_loss + actor_loss
        return loss, info

    def target_update(self, network, module_name):
        """Update the target network."""
        new_target_params = jax.tree_util.tree_map(
            lambda p, tp: p * self.config['tau'] + tp * (1 - self.config['tau']),
            self.network.params[f'modules_{module_name}'],
            self.network.params[f'modules_target_{module_name}'],
        )
        network.params[f'modules_target_{module_name}'] = new_target_params

    @jax.jit
    def update(self, batch):
        """Update the agent and return a new agent with information dictionary."""
        new_rng, rng = jax.random.split(self.rng)

        def loss_fn(grad_params):
            return self.total_loss(batch, grad_params, rng=rng)

        new_network, info = self.network.apply_loss_fn(loss_fn=loss_fn)
        self.target_update(new_network, 'value')

        return self.replace(network=new_network, rng=new_rng), info

    @jax.jit
    def sample_actions(
        self,
        observations,
        goals=None,
        seed=None,
        temperature=1.0,
    ):
        """Sample actions from the actor."""
        dist = self.network.select('actor')(observations, goals, temperature=temperature)
        actions = dist.sample(seed=seed)
        if not self.config['discrete']:
            actions = jnp.clip(actions, -1, 1)
        return actions

    @classmethod
    def create(
        cls,
        seed,
        ex_observations,
        ex_actions,
        config,
    ):
        """Create a new agent.

        Args:
            seed: Random seed.
            ex_observations: Example batch of observations.
            ex_actions: Example batch of actions. In discrete-action MDPs, this should contain the maximum action value.
            config: Configuration dictionary.
        """
        rng = jax.random.PRNGKey(seed)
        rng, init_rng = jax.random.split(rng, 2)

        ex_goals = ex_observations
        if config['discrete']:
            action_dim = ex_actions.max() + 1
        else:
            action_dim = ex_actions.shape[-1]

        # Define encoders.
        encoders = dict()
        if config['encoder'] is not None:
            encoder_module = encoder_modules[config['encoder']]
            encoders['value'] = GCEncoder(concat_encoder=encoder_module())
            encoders['actor'] = GCEncoder(concat_encoder=encoder_module())

        # Define value and actor networks.
        value_def = GCValue(
            hidden_dims=config['value_hidden_dims'],
            layer_norm=config['layer_norm'],
            ensemble=True,
            gc_encoder=encoders.get('value'),
        )

        if config['discrete']:
            actor_def = GCDiscreteActor(
                hidden_dims=config['actor_hidden_dims'],
                action_dim=action_dim,
                gc_encoder=encoders.get('actor'),
            )
        else:
            actor_def = GCActor(
                hidden_dims=config['actor_hidden_dims'],
                action_dim=action_dim,
                state_dependent_std=False,
                const_std=config['const_std'],
                gc_encoder=encoders.get('actor'),
            )

        network_info = dict(
            value=(value_def, (ex_observations, ex_goals)),
            target_value=(copy.deepcopy(value_def), (ex_observations, ex_goals)),
            actor=(actor_def, (ex_observations, ex_goals)),
        )
        networks = {k: v[0] for k, v in network_info.items()}
        network_args = {k: v[1] for k, v in network_info.items()}

        network_def = ModuleDict(networks)
        network_tx = optax.adam(learning_rate=config['lr'])
        network_params = network_def.init(init_rng, **network_args)['params']
        network = TrainState.create(network_def, network_params, tx=network_tx)

        params = network_params
        params['modules_target_value'] = params['modules_value']

        return cls(rng, network=network, config=flax.core.FrozenDict(**config))


def get_config():
    config = ml_collections.ConfigDict(
        dict(
            # Agent hyperparameters.
            agent_name='gcivl',  # Agent name.
            lr=3e-4,  # Learning rate.
            batch_size=1024,  # Batch size.
            actor_hidden_dims=(512, 512, 512),  # Actor network hidden dimensions.
            value_hidden_dims=(512, 512, 512),  # Value network hidden dimensions.
            layer_norm=True,  # Whether to use layer normalization.
            discount=0.99,  # Discount factor.
            tau=0.005,  # Target network update rate.
            expectile=0.9,  # IQL expectile.
            alpha=10.0,  # AWR temperature.
            const_std=True,  # Whether to use constant standard deviation for the actor.
            discrete=False,  # Whether the action space is discrete.
            encoder=ml_collections.config_dict.placeholder(str),  # Visual encoder name (None, 'impala_small', etc.).
            # Dataset hyperparameters.
            dataset_class='GCDataset',  # Dataset class name.
            value_p_curgoal=0.2,  # Probability of using the current state as the value goal.
            value_p_trajgoal=0.5,  # Probability of using a future state in the same trajectory as the value goal.
            value_p_randomgoal=0.3,  # Probability of using a random state as the value goal.
            value_geom_sample=True,  # Whether to use geometric sampling for future value goals.
            actor_p_curgoal=0.0,  # Probability of using the current state as the actor goal.
            actor_p_trajgoal=1.0,  # Probability of using a future state in the same trajectory as the actor goal.
            actor_p_randomgoal=0.0,  # Probability of using a random state as the actor goal.
            actor_geom_sample=False,  # Whether to use geometric sampling for future actor goals.
            gc_negative=True,  # Whether to use '0 if s == g else -1' (True) or '1 if s == g else 0' (False) as reward.
            p_aug=0.0,  # Probability of applying image augmentation.
            frame_stack=ml_collections.config_dict.placeholder(int),  # Number of frames to stack.
        )
    )
    return config



import gymnasium
import numpy as np
from gymnasium.spaces import Box, Discrete
from PIL import Image

from ogbench.powderworld.sim import PWRenderer, PWSim, interp, pw_element_names


class PowderworldEnv(gymnasium.Env):
    """Powderworld enviroment.

    This wraps the numpy version of the Powderworld simulator to create a gymnasium environment. It supports
    goal-based tasks where the agent must recreate a given world state.
    """

    metadata = {
        'render_modes': ['rgb_array'],
        'render_fps': 15,
    }

    def __init__(
        self,
        world_size=32,
        grid_size=4,
        brush_size=4,
        num_elems=5,
        mode='task',
        render_mode=None,
        width=192,
        height=192,
    ):
        """Initialize the Powderworld environment.

        Args:
            world_size: Size of the world grid.
            grid_size: Size of the action grid.
            brush_size: Size of the brush.
            num_elems: Number of elements in the world. Must be 2, 5, or 8.
            mode: Mode of the environment. Either 'task' or 'data_collection'. In 'task' mode, the environment is used
                for training and evaluation. In 'data_collection' mode, the environment is used for collecting offline
                data.
            render_mode: Rendering mode. Unused; for compatibility with `gymnasium`.
            width: Width of the rendered image. Only used for rendering, not for observations.
            height: Height of the rendered image. Only used for rendering, not for observations.
        """
        self.pw = PWSim()
        self.pwr = PWRenderer()

        self._world_size = world_size
        self._grid_size = grid_size
        self._brush_size = brush_size
        self._mode = mode
        self._render_width = width
        self._render_height = height
        self._num_elems = num_elems
        if num_elems == 2:
            self._elem_names = ['plant', 'stone']
        elif num_elems == 5:
            self._elem_names = ['sand', 'water', 'fire', 'plant', 'stone']
        elif num_elems == 8:
            self._elem_names = ['sand', 'water', 'fire', 'plant', 'stone', 'gas', 'wood', 'ice']
        else:
            raise NotImplementedError
        self._elems = [pw_element_names.index(elem_name) for elem_name in self._elem_names]
        self._elem_colors = self.pwr.elem_vecs_array[self._elems].copy()

        self.observation_space = Box(low=0, high=255, shape=(self._world_size, self._world_size, 6), dtype=np.uint8)
        self._xy_action_size = (world_size - brush_size) // grid_size + 1
        self.action_space = Discrete(max(len(self._elems), self._xy_action_size))

        self._world = None

        # The environment has three action stages:
        # 0: Set element id.
        # 1: Set the x coordinate.
        # 2: Set the y coordinate.
        self._action_step = 0  # The current stage.
        self._action_elem_id = None  # The selected element id.
        self._action_x = None  # The selected x coordinate.
        self._action_y = None  # The selected y coordinate

        if self._mode == 'task':
            # Set task goals.
            self.task_infos = []
            self.cur_task_id = None
            self.cur_task_info = None
            self.set_tasks()
            self.num_tasks = len(self.task_infos)
            self.cur_goal_world = None

    def set_tasks(self):
        def add_square(action_seq, elem_name, x, y, size):
            """Add a square to the action sequence."""
            for i in range(0, size):
                action_seq.append((elem_name, x + i, y + size - 1))
            for i in range(size - 2, -1, -1):
                action_seq.append((elem_name, x, y + i))
            for i in range(size - 2, -1, -1):
                action_seq.append((elem_name, x + size - 1, y + i))
            for i in range(1, size - 1):
                action_seq.append((elem_name, x + i, y))

        if self._num_elems == 2:
            # Task 1: Plant.
            action_seq = []
            for y in reversed(range(8)):
                for x in range(8):
                    action_seq.append(('plant', x, y))
            self.task_infos.append(dict(task_name='task1_plant', action_seq=action_seq, tol=32))

            # Task 2: Stone.
            action_seq = []
            for y in reversed(range(8)):
                for x in range(8):
                    action_seq.append(('plant', x, y))
            for y in reversed(range(8)):
                for x in range(8):
                    action_seq.append(('stone', x, y))
            self.task_infos.append(dict(task_name='task2_stone', action_seq=action_seq, tol=32))

            # Task 3: Square.
            action_seq = []
            for y in reversed(range(8)):
                for x in range(8):
                    action_seq.append(('plant', x, y))
            add_square(action_seq, 'stone', 1, 1, 6)
            self.task_infos.append(dict(task_name='task3_square', action_seq=action_seq, tol=32))

            # Task 4: Four squares.
            action_seq = []
            for y in reversed(range(8)):
                for x in range(8):
                    action_seq.append(('plant', x, y))
            for y in reversed(range(8)):
                for x in range(8):
                    action_seq.append(('stone', x, y))
            for sx, sy in [(0, 0), (0, 5), (5, 0), (5, 5)]:
                add_square(action_seq, 'plant', sx, sy, 3)
            self.task_infos.append(dict(task_name='task4_four_squares', action_seq=action_seq, tol=32))

            # Task 5: Mosaic.
            action_seq = []
            for y in reversed(range(8)):
                for x in range(8):
                    action_seq.append(('plant', x, y))
            for y in reversed(range(8)):
                for x in range(8):
                    if (x + y) % 2 == 0:
                        action_seq.append(('stone', x, y))
            self.task_infos.append(dict(task_name='task5_mosaic', action_seq=action_seq, tol=32))
        elif self._num_elems == 5:
            # Task 1: Squares.
            action_seq = []
            add_square(action_seq, 'plant', 1, 1, 6)
            add_square(action_seq, 'sand', 0, 0, 8)
            add_square(action_seq, 'stone', 2, 2, 4)
            add_square(action_seq, 'water', 3, 3, 2)
            self.task_infos.append(dict(task_name='task1_squares', action_seq=action_seq, tol=32))

            # Task 2: Water plant.
            action_seq = []
            for y in reversed(range(8)):
                for x in range(8):
                    action_seq.append(('water', x, y))
            add_square(action_seq, 'plant', 0, 0, 8)
            self.task_infos.append(dict(task_name='task2_water_plant', action_seq=action_seq, tol=64))

            # Task 3: Sandpile.
            action_seq = []
            add_square(action_seq, 'stone', 0, 0, 8)
            for _ in range(32):
                action_seq.extend(
                    [
                        ('sand', 3, 1),
                        ('sand', 4, 1),
                    ]
                )
            self.task_infos.append(dict(task_name='task3_sandpile', action_seq=action_seq, tol=64))

            # Task 4: Two rooms.
            action_seq = []
            for x in range(0, 8):
                action_seq.append(('plant', x, 6))
            for x in range(0, 8):
                action_seq.append(('plant', x, 7))
            for y in range(7, -1, -1):
                action_seq.append(('stone', 0, y))
            for y in range(7, -1, -1):
                action_seq.append(('stone', 7, y))
            for x in range(1, 7):
                action_seq.append(('stone', x, 4))
            for x in range(1, 7):
                action_seq.append(('stone', x, 3))
            add_square(action_seq, 'water', 2, 0, 3)
            add_square(action_seq, 'water', 3, 0, 3)
            for _ in range(4):
                action_seq.append(('fire', 3, 7))
                action_seq.append(('fire', 4, 7))
            self.task_infos.append(dict(task_name='task4_two_rooms', action_seq=action_seq, tol=64))

            # Task 5: Elements.
            action_seq = []
            for y in reversed(range(8)):
                for x in range(8):
                    action_seq.append(('plant', x, y))
            for y in [4, 7]:
                for x in range(8):
                    action_seq.append(('water', x, y))
            for _ in range(2):
                for x in range(8):
                    action_seq.append(('fire', x, 0))
            self.task_infos.append(dict(task_name='task5_elements', action_seq=action_seq, tol=96))
        elif self._num_elems == 8:
            # Task 1: Bubbles.
            action_seq = []
            for y in range(7, -1, -1):
                for x in range(8):
                    action_seq.append(('sand', x, y))
            for x in range(8):
                action_seq.append(('sand', x, 0))
            for x in range(8):
                action_seq.append(('water', x, 7))
            for x in range(8):
                action_seq.append(('gas', x, 7))
            for x in range(8):
                action_seq.append(('water', x, 7))
            self.task_infos.append(dict(task_name='task1_bubbles', action_seq=action_seq, tol=96))

            # Task 2: Firework.
            action_seq = []
            add_square(action_seq, 'wood', 0, 0, 8)
            add_square(action_seq, 'plant', 1, 1, 6)
            add_square(action_seq, 'gas', 2, 2, 4)
            for _ in range(3):
                for x in range(8):
                    action_seq.append(('fire', x, 0))
            self.task_infos.append(dict(task_name='task2_firework', action_seq=action_seq, tol=96))

            # Task 3: Three rooms.
            action_seq = []
            for x in range(8):
                action_seq.append(('ice', x, 0))
            for y in range(7, -1, -1):
                action_seq.append(('stone', 2, y))
            for y in range(7, -1, -1):
                action_seq.append(('stone', 5, y))
            for y in range(7, 0, -1):
                action_seq.append(('water', 3, y))
                action_seq.append(('water', 4, y))
            action_seq.append(('plant', 3, 3))
            action_seq.append(('plant', 4, 3))
            action_seq.append(('plant', 3, 4))
            action_seq.append(('plant', 4, 4))
            for y in range(7, 0, -1):
                action_seq.append(('gas', 0, y))
                action_seq.append(('gas', 1, y))
                action_seq.append(('gas', 6, y))
                action_seq.append(('gas', 7, y))
            self.task_infos.append(dict(task_name='task3_three_rooms', action_seq=action_seq, tol=96))

            # Task 4: Four squares.
            action_seq = []
            add_square(action_seq, 'plant', 1, 4, 3)
            add_square(action_seq, 'wood', 4, 4, 3)
            add_square(action_seq, 'ice', 1, 1, 3)
            add_square(action_seq, 'plant', 4, 1, 3)
            for _ in range(10):
                add_square(action_seq, 'plant', 4, 1, 3)
            self.task_infos.append(dict(task_name='task4_four_squares', action_seq=action_seq, tol=64))

            # Task 5: Ice plant.
            action_seq = []
            for y in range(7, -1, -1):
                for x in range(8):
                    action_seq.append(('water', x, y))
            add_square(action_seq, 'plant', 3, 3, 2)
            for _ in range(4):
                add_square(action_seq, 'stone', 0, 0, 8)
            add_square(action_seq, 'ice', 3, 3, 2)
            self.task_infos.append(dict(task_name='task5_ice_plant', action_seq=action_seq, tol=96))
        else:
            raise NotImplementedError

    def reset(self, *, seed=None, options=None):
        if self._mode == 'task':
            # Set the task goal.
            if options is None:
                options = {}

            if 'task_id' in options:
                # Use the pre-defined task.
                assert 1 <= options['task_id'] <= self.num_tasks, f'Task ID must be in [1, {self.num_tasks}].'
                self.cur_task_id = options['task_id']
                self.cur_task_info = self.task_infos[self.cur_task_id - 1]
            elif 'task_info' in options:
                # Use the provided task information.
                self.cur_task_id = None
                self.cur_task_info = options['task_info']
            else:
                # Randomly sample a task.
                self.cur_task_id = np.random.randint(1, self.num_tasks + 1)
                self.cur_task_info = self.task_infos[self.cur_task_id - 1]

            # Whether to provide a rendering of the goal.
            render_goal = False
            if 'render_goal' in options:
                render_goal = options['render_goal']

        # Initialize a new world.
        np_world = np.zeros((1, self._world_size, self._world_size), dtype=np.uint8)
        np_world[:, 0, :] = 1
        np_world[:, -1, :] = 1
        np_world[:, :, 0] = 1
        np_world[:, :, -1] = 1

        self._world = self.pw.np_to_pw(np_world).copy()
        self._action_step = 0
        self._action_elem_id = None
        self._action_x = None
        self._action_y = None

        if self._mode == 'task':
            # Before doing the actual reset, we simulate the goal actions to get a goal observation.
            self._mode = 'internal'  # Set the mode to 'internal' to prevent the agent from computing the success.
            self.reset()
            for semantic_action in self.cur_task_info['action_seq']:
                for _ in range(3):
                    self.step(self.semantic_action_to_action(*semantic_action))
            self._mode = 'task'
            goal = self._get_ob()
            self.cur_goal_world = self._world[0, 0].copy()
            if render_goal:
                goal_rendered = self.render()

            # Do the actual reset.
            self._mode = 'internal'  # Set the mode to 'internal' to prevent the agent from computing the success.
            self.reset()
            semantic_action = self.sample_semantic_action()
            # Perform a single random action (3 steps) to add some randomness to the initial state.
            for _ in range(3):
                self.step(self.semantic_action_to_action(*semantic_action))
            self._mode = 'task'

        ob = self._get_ob()
        info = dict()

        if self._mode == 'task':
            info['goal'] = goal
            if render_goal:
                info['goal_rendered'] = goal_rendered

        return ob, info

    def step(self, action):
        if self._action_step == 0:
            # Step 0: Set the element id.
            if action < len(self._elems):
                self._action_elem_id = action
            else:
                # If the action is invalid, randomly select an element.
                self._action_elem_id = np.random.randint(len(self._elems))
        elif self._action_step == 1:
            # Step 1: Set the x coordinate.
            if action < self._xy_action_size:
                self._action_x = action
            else:
                # If the action is invalid, randomly select an x coordinate.
                self._action_x = np.random.randint(self._xy_action_size)
        else:
            # Step 2: Set the y coordinate.
            if action < self._xy_action_size:
                self._action_y = action
            else:
                # If the action is invalid, randomly select a y coordinate.
                self._action_y = np.random.randint(self._xy_action_size)

            # Step the world.
            self._world = self.pw.forward(self._world)

            # Apply the action.
            np_action_world = np.zeros((1, self._world_size, self._world_size), dtype=np.uint8)
            elem = self._elems[self._action_elem_id]
            real_x = self._action_x * self._grid_size
            real_y = self._action_y * self._grid_size
            np_action_world[:, real_y : real_y + self._brush_size, real_x : real_x + self._brush_size] = elem
            world_delta = self.pw.np_to_pw(np_action_world)
            self._world = interp(
                ~self.pw.get_bool(world_delta, 'empty') & ~self.pw.get_bool(self._world, 'wall'),
                self._world,
                world_delta,
            )

            # Reset the action step.
            self._action_elem_id = None
            self._action_x = None
            self._action_y = None

        self._action_step = (self._action_step + 1) % 3

        ob = self._get_ob()
        reward = 0.0
        terminated = False
        info = dict()

        if self._mode == 'task':
            # Check if the current world matches the goal. To have some tolerance, for each pixel in the goal, we check
            # if the current world has a matching pixel that is shifted by up to one pixel in any direction. We then
            # compute the error as the number of pixels that do not match, and consider the task successful if the error
            # is below a certain threshold.
            cur_world = self._world[0, 0].copy()
            world_shifts = []
            for dx, dy in [(0, 0), (1, 0), (-1, 0), (0, 1), (0, -1)]:
                world_shifts.append(np.roll(cur_world, (dy, dx), axis=(0, 1)))
            world_shifts = np.stack(world_shifts, axis=0)
            match = (self.cur_goal_world == world_shifts).any(axis=0)
            error = ~match

            success = error.sum() < self.cur_task_info['tol']
            if success:
                terminated = True
                info['success'] = 1.0
                reward = 1.0
            else:
                info['success'] = 0.0
                reward = 0.0

        return ob, reward, terminated, False, info

    def semantic_action_to_action(self, elem_name, x, y):
        """Convert a semantic action to an action based on the current action step."""
        elem_id = self._elem_names.index(elem_name)
        if self._action_step == 0:
            return elem_id
        elif self._action_step == 1:
            return x
        else:
            return y

    def sample_action(self):
        """Sample a random valid action."""
        if self._action_step == 0:
            return np.random.randint(len(self._elems))
        else:
            return np.random.randint(self._xy_action_size)

    def sample_semantic_action(self):
        """Sample a random semantic action."""
        elem_name = np.random.choice(self._elem_names)
        x = np.random.randint(self._xy_action_size)
        y = np.random.randint(self._xy_action_size)
        return elem_name, x, y

    def render(self):
        ob = self._get_ob()
        frame = ob[..., :3]
        frame = Image.fromarray(frame)
        frame = frame.resize((self._render_width, self._render_height), Image.NEAREST)
        frame = np.array(frame)

        return frame

    def _get_ob(self):
        world_frame = self.pwr.render(self._world).copy()  # (world_size, world_size, 3)

        # Compute an additional action frame to show the currently selected action.
        action_frame = np.zeros_like(world_frame)  # (world_size, world_size, 3)
        if self._action_step == 1:
            # Color the entire frame with the selected element color.
            color = self._elem_colors[self._action_elem_id]
            action_frame[..., :] = (color * 255.0).astype(np.uint8)
        elif self._action_step == 2:
            # Color the selected x coordinate with the selected element color.
            color = self._elem_colors[self._action_elem_id]
            real_x = self._action_x * self._grid_size
            action_frame[:, real_x : real_x + self._brush_size, :] = (color * 255.0).astype(np.uint8)
        return np.concatenate([world_frame, action_frame], axis=2)  # (world_size, world_size, 6)



from agents.crl import CRLAgent
from agents.gcbc import GCBCAgent
from agents.gciql import GCIQLAgent
from agents.gcivl import GCIVLAgent
from agents.hiql import HIQLAgent
from agents.qrl import QRLAgent
from agents.sac import SACAgent

agents = dict(
    crl=CRLAgent,
    gcbc=GCBCAgent,
    gciql=GCIQLAgent,
    gcivl=GCIVLAgent,
    hiql=HIQLAgent,
    qrl=QRLAgent,
    sac=SACAgent,
)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import mujoco
import numpy as np

from ogbench.manipspace.lie.utils import get_epsilon

_IDENTITIY_WXYZ = np.array([1.0, 0.0, 0.0, 0.0], dtype=np.float64)
_INVERT_QUAT_SIGN = np.array([1.0, -1.0, -1.0, -1.0], dtype=np.float64)


@dataclass(frozen=True)
class RollPitchYaw:
    roll: float
    pitch: float
    yaw: float


@dataclass(frozen=True)
class SO3:
    """Special orthogonal group for 3D rotations.

    Internal parameterization is (qw, qx, qy, qz).
    Tangent parameterization is (omega_x, omega_y, omega_z).
    """

    wxyz: np.ndarray
    matrix_dim: int = 3
    parameters_dim: int = 4
    tangent_dim: int = 3
    space_dim: int = 3

    def __post_init__(self) -> None:
        if self.wxyz.shape != (self.parameters_dim,):
            raise ValueError(f'Expeced wxyz to be a length 4 vector but got {self.wxyz.shape[0]}.')

    def __repr__(self) -> str:
        wxyz = np.round(self.wxyz, 5)
        return f'{self.__class__.__name__}(wxyz={wxyz})'

    def copy(self) -> SO3:
        return SO3(wxyz=self.wxyz.copy())

    @staticmethod
    def from_x_radians(theta: float) -> SO3:
        return SO3.exp(np.array([theta, 0.0, 0.0], dtype=np.float64))

    @staticmethod
    def from_y_radians(theta: float) -> SO3:
        return SO3.exp(np.array([0.0, theta, 0.0], dtype=np.float64))

    @staticmethod
    def from_z_radians(theta: float) -> SO3:
        return SO3.exp(np.array([0.0, 0.0, theta], dtype=np.float64))

    @staticmethod
    def from_rpy_radians(
        roll: float,
        pitch: float,
        yaw: float,
    ) -> SO3:
        return SO3.from_z_radians(yaw) @ SO3.from_y_radians(pitch) @ SO3.from_x_radians(roll)

    @staticmethod
    def from_matrix(matrix: np.ndarray) -> SO3:
        assert matrix.shape == (SO3.matrix_dim, SO3.matrix_dim)
        wxyz = np.zeros(SO3.parameters_dim, dtype=np.float64)
        mujoco.mju_mat2Quat(wxyz, matrix.ravel())
        return SO3(wxyz=wxyz)

    @staticmethod
    def identity() -> SO3:
        return SO3(wxyz=_IDENTITIY_WXYZ)

    @staticmethod
    def sample_uniform() -> SO3:
        u1, u2, u3 = np.random.uniform(
            low=np.zeros(shape=(3,)),
            high=np.array([1.0, 2.0 * np.pi, 2.0 * np.pi]),
        )
        a = np.sqrt(1.0 - u1)
        b = np.sqrt(u1)
        wxyz = np.array(
            [
                a * np.sin(u2),
                a * np.cos(u2),
                b * np.sin(u3),
                b * np.cos(u3),
            ],
            dtype=np.float64,
        )
        return SO3(wxyz=wxyz)

    def as_matrix(self) -> np.ndarray:
        mat = np.zeros(9, dtype=np.float64)
        mujoco.mju_quat2Mat(mat, self.wxyz)
        return mat.reshape(3, 3)

    def compute_roll_radians(self) -> float:
        q0, q1, q2, q3 = self.wxyz
        return np.arctan2(2 * (q0 * q1 + q2 * q3), 1 - 2 * (q1**2 + q2**2))

    def compute_pitch_radians(self) -> float:
        q0, q1, q2, q3 = self.wxyz
        return np.arcsin(2 * (q0 * q2 - q3 * q1))

    def compute_yaw_radians(self) -> float:
        q0, q1, q2, q3 = self.wxyz
        return np.arctan2(2 * (q0 * q3 + q1 * q2), 1 - 2 * (q2**2 + q3**2))

    def as_rpy_radians(self) -> RollPitchYaw:
        return RollPitchYaw(
            roll=self.compute_roll_radians(),
            pitch=self.compute_pitch_radians(),
            yaw=self.compute_yaw_radians(),
        )

    @staticmethod
    def exp(tangent: np.ndarray) -> SO3:
        assert tangent.shape == (SO3.tangent_dim,)
        theta_squared = tangent @ tangent
        theta_pow_4 = theta_squared * theta_squared
        use_taylor = theta_squared < get_epsilon(tangent.dtype)
        safe_theta = 1.0 if use_taylor else np.sqrt(theta_squared)
        safe_half_theta = 0.5 * safe_theta
        if use_taylor:
            real = 1.0 - theta_squared / 8.0 + theta_pow_4 / 384.0
            imaginary = 0.5 - theta_squared / 48.0 + theta_pow_4 / 3840.0
        else:
            real = np.cos(safe_half_theta)
            imaginary = np.sin(safe_half_theta) / safe_theta
        wxyz = np.concatenate([np.array([real]), imaginary * tangent])
        return SO3(wxyz=wxyz)

    def log(self) -> np.ndarray:
        w = self.wxyz[0]
        norm_sq = self.wxyz[1:] @ self.wxyz[1:]
        use_taylor = norm_sq < get_epsilon(norm_sq.dtype)
        norm_safe = 1.0 if use_taylor else np.sqrt(norm_sq)
        w_safe = w if use_taylor else 1.0
        atan_n_over_w = np.arctan2(-norm_safe if w < 0 else norm_safe, abs(w))
        if use_taylor:
            atan_factor = 2.0 / w_safe - 2.0 / 3.0 * norm_sq / w_safe**3
        else:
            if abs(w) < get_epsilon(w.dtype):
                scl = 1.0 if w > 0.0 else -1.0
                atan_factor = scl * np.pi / norm_safe
            else:
                atan_factor = 2.0 * atan_n_over_w / norm_safe
        return atan_factor * self.wxyz[1:]

    def adjoint(self) -> np.ndarray:
        return self.as_matrix()

    def inverse(self) -> SO3:
        return SO3(wxyz=self.wxyz * _INVERT_QUAT_SIGN)

    def normalize(self) -> SO3:
        return SO3(wxyz=self.wxyz / np.linalg.norm(self.wxyz))

    def apply(self, target: np.ndarray) -> np.ndarray:
        assert target.shape == (SO3.space_dim,)
        padded_target = np.concatenate([np.zeros(1, dtype=np.float64), target])
        return (self @ SO3(wxyz=padded_target) @ self.inverse()).wxyz[1:]

    def multiply(self, other: SO3) -> SO3:
        w0, x0, y0, z0 = self.wxyz
        w1, x1, y1, z1 = other.wxyz
        return SO3(
            wxyz=np.array(
                [
                    -x0 * x1 - y0 * y1 - z0 * z1 + w0 * w1,
                    x0 * w1 + y0 * z1 - z0 * y1 + w0 * x1,
                    -x0 * z1 + y0 * w1 + z0 * x1 + w0 * y1,
                    x0 * y1 - y0 * x1 + z0 * w1 + w0 * z1,
                ],
                dtype=np.float64,
            )
        )

    def __matmul__(self, other: Any) -> Any:
        if isinstance(other, np.ndarray):
            return self.apply(target=other)
        elif isinstance(other, SO3):
            return self.multiply(other=other)
        else:
            raise ValueError(f'Unsupported argument type for @ operator: {type(other)}')
